"""
Tests for ``nefelibata.assistants.archive_links``.
"""

from pathlib import Path

import pytest
from pytest_mock import MockerFixture
from yarl import URL

from nefelibata.assistants.archive_links import ArchiveLinksAssistant
from nefelibata.config import Config
from nefelibata.post import Post


@pytest.mark.asyncio
async def test_assistant(
    mocker: MockerFixture,
    root: Path,
    config: Config,
    post: Post,
) -> None:
    """
    Test the assistant.
    """
    mocker.patch(
        "nefelibata.assistants.archive_links.archive_urls",
        return_value={
            URL("https://nefelibata.readthedocs.io/"): URL(
                "https://web.archive.org/web/20211003154602/https://nefelibata.readthedocs.io/",
            ),
        },
    )

    assistant = ArchiveLinksAssistant(root, config)

    response = await assistant.get_post_metadata(post)
    assert response == {
        "https://nefelibata.readthedocs.io/": (
            "https://web.archive.org/web/20211003154602/"
            "https://nefelibata.readthedocs.io/"
        ),
    }



"""
Build the blog.
"""
import logging
import os
from pathlib import Path
from subprocess import call

from werkzeug.utils import secure_filename

from nefelibata.utils import get_config

_logger = logging.getLogger(__name__)


async def run(root: Path, title: str, type_: str = "post") -> None:
    """
    Create a new post and open editor.
    """
    _logger.info("Creating new directory")

    directory = secure_filename(title).lower()
    target = root / "posts" / directory
    if target.exists():
        raise IOError("Directory already exists!")
    target.mkdir(parents=True)

    headers = {
        "subject": title,
        "summary": "",
        "keywords": "",
    }
    if type_ != "post":
        config = get_config(root)
        try:
            extra_headers = config.templates[type_]
        except KeyError as ex:
            raise Exception(f"Invalid post type: {type_}") from ex

        headers["type"] = type_
        headers.update({f"{type_}-{key}": "" for key in extra_headers})

    filepath = target / "index.mkd"
    with open(filepath, "w", encoding="utf-8") as output:
        for key, value in headers.items():
            output.write(f"{key}: {value}\n")
        output.write("\n\n")

    editor = os.environ.get("EDITOR")
    if not editor:
        _logger.info("No EDITOR found, exiting")
        return

    call([editor, filepath])



"""
An S3 publisher.
"""
import logging
import mimetypes
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Optional

import boto3
from botocore.exceptions import ClientError

from nefelibata.config import Config
from nefelibata.publishers.base import Publisher, Publishing

_logger = logging.getLogger(__name__)


class S3Publisher(Publisher):

    """
    A publisher that uploads the blog to an S3 service
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        root: Path,
        config: Config,
        path: str,
        access_key_id: str,
        secret_access_key: str,
        bucket: str,
        region: str = "us-east-1",
        **kwargs: Any,
    ):
        super().__init__(root, config, path, **kwargs)

        self.bucket = bucket
        self.client = boto3.client(
            "s3",
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
            region_name=region,
        )

    async def publish(
        self,
        since: Optional[datetime] = None,
        force: bool = False,
    ) -> Optional[Publishing]:
        build = self.root / "build" / self.path

        modified_files = list(self.find_modified_files(force, since))

        for path in modified_files:
            key = str(path.relative_to(build))
            self._upload_file(path, key)

        if modified_files:
            return Publishing(timestamp=datetime.now(timezone.utc))

        return None

    def _upload_file(self, path: Path, key: str) -> None:
        """
        Upload a file to S3.
        """
        extra_args = {"ACL": "public-read"}
        mimetype = mimetypes.guess_type(str(path))[0]
        if mimetype:
            extra_args["ContentType"] = mimetype

        _logger.info("Uploading %s to %s in S3", path, self.bucket)
        try:
            self.client.upload_file(str(path), self.bucket, key, ExtraArgs=extra_args)
        except ClientError as ex:
            _logger.exception("An error occurred")
            raise ex



"""
Nefelibata blog engine.

Usage:
  nb init [ROOT_DIR] [-f] [--loglevel=INFO]
  nb new POST [ROOT_DIR] [-t TYPE] [--loglevel=INFO]
  nb build [ROOT_DIR] [-f] [--loglevel=INFO]
  nb publish [ROOT_DIR] [-f] [--loglevel=INFO]

Actions:
  init              Create a new blog skeleton.
  new               Create a new post.
  build             Build blog from Markdown files and online interactions.
  publish           Publish weblog to configured locations.

Options:
  -h --help         Show this screen.
  --version         Show version.
  -f --force        Force operation (eg, building up-to-date resources).
  -t TYPE           Custom template to use on the post. [default: post]
  --loglevel=LEVEL  Level for logging. [default: INFO]

Released under the MIT license.
(c) 2013 Beto Dealmeida <roberto@dealmeida.net>
"""
import asyncio
import logging
import os
from pathlib import Path

from docopt import docopt

from nefelibata import __version__
from nefelibata.cli import build, init, new, publish
from nefelibata.utils import find_directory, setup_logging

_logger = logging.getLogger(__name__)


async def main() -> None:
    """
    Dispatch command.
    """
    arguments = docopt(__doc__, version=__version__)

    setup_logging(arguments["--loglevel"])

    if arguments["ROOT_DIR"] is None:
        root = find_directory(Path(os.getcwd()))
    else:
        root = Path(arguments["ROOT_DIR"])

    try:
        if arguments["init"]:
            await init.run(root, arguments["--force"])
        elif arguments["new"]:
            await new.run(root, arguments["POST"], arguments["-t"])
        elif arguments["build"]:
            await build.run(root, arguments["--force"])
        elif arguments["publish"]:
            await publish.run(root, arguments["--force"])
    except asyncio.CancelledError:
        _logger.info("Canceled")


def run() -> None:
    """
    Run Nefelibata.

    This is also an entry point for the CLI.
    """
    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        _logger.info("Stopping Nefelibata")


if __name__ == "__main__":
    run()



"""
Tests for ``nefelibata.post``.
"""
# pylint: disable=invalid-name, unused-argument

from datetime import datetime, timezone
from pathlib import Path

import pytest
from freezegun import freeze_time
from pyfakefs.fake_filesystem import FakeFilesystem
from yarl import URL

from nefelibata.config import AnnouncerModel, Config
from nefelibata.post import Post, build_post, extract_links, get_posts

from .fakes import POST_CONTENT, POST_DATA


@pytest.mark.asyncio
async def test_post() -> None:
    """
    Test the ``Post`` model.
    """
    post = Post(**POST_DATA)

    assert post.path == Path("/path/to/blog/posts/first/index.mkd")
    assert post.title == "This is your first post"
    assert post.timestamp == datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc)
    assert post.metadata == {
        "keywords": "welcome, blog",
        "summary": "Hello, world!",
    }
    assert post.tags == {"welcome", "blog"}
    assert post.categories == {"stem"}
    assert post.type == "post"
    assert post.url == "first/index"
    assert (
        post.content
        == """# Welcome #

This is your first post. It should be written using Markdown.

Read more about [Nefelibata](https://nefelibata.readthedocs.io/)."""
    )


@pytest.mark.asyncio
async def test_build_post(fs: FakeFilesystem, root: Path, config: Config) -> None:
    """
    Test ``build_post``.
    """
    path = Path(root / "posts/first/index.mkd")

    # create post
    with freeze_time("2021-01-01T00:00:00Z"):
        fs.create_file(path, contents=POST_CONTENT)

    config.announcers = {"antenna": AnnouncerModel(plugin="antenna")}

    post = build_post(root, config, path)

    assert post.path == path
    assert post.title == "This is your first post"
    assert post.timestamp == datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc)
    assert post.metadata == {
        "keywords": "welcome, blog",
        "summary": "Hello, world!",
    }
    assert post.tags == {"welcome", "blog"}
    assert post.categories == {"stem"}
    assert post.type == "post"
    assert post.url == "first/index"
    assert (
        post.content
        == """# Welcome #

This is your first post. It should be written using Markdown.

Read more about [Nefelibata](https://nefelibata.readthedocs.io/)."""
    )
    assert post.announcers == {"antenna"}

    # check to file was updated with the ``date`` header
    with open(path, encoding="utf-8") as input_:
        content = input_.read()
    assert (
        content
        == """subject: This is your first post
keywords: welcome, blog
summary: Hello, world!
date: Thu, 31 Dec 2020 16:00:00 -0800
announce-on: antenna

# Welcome #

This is your first post. It should be written using Markdown.

Read more about [Nefelibata](https://nefelibata.readthedocs.io/)."""
    )

    # build again, and test that the file wasn't modified since it
    # already has all the required headers
    last_update = path.stat().st_mtime
    build_post(root, config, path)
    assert path.stat().st_mtime == last_update


@pytest.mark.asyncio
async def test_get_posts(fs: FakeFilesystem, root: Path, config: Config) -> None:
    """
    Test ``get_posts``.
    """
    with freeze_time("2021-01-01T00:00:00Z"):
        fs.create_dir(root / "posts/one")
        fs.create_file(root / "posts/one/index.mkd")
    with freeze_time("2021-01-02T00:00:00Z"):
        fs.create_dir(root / "posts/two")
        fs.create_file(root / "posts/two/index.mkd")

    posts = get_posts(root, config)
    assert len(posts) == 2
    assert posts[0].path == Path(root / "posts/two/index.mkd")
    assert posts[1].path == Path(root / "posts/one/index.mkd")

    # test limited number of posts returned
    posts = get_posts(root, config, 1)
    assert len(posts) == 1


def test_extract_links(post: Post) -> None:
    """
    Test ``extract_links``.
    """
    assert list(extract_links(post)) == [URL("https://nefelibata.readthedocs.io/")]

    post.metadata["test-url"] = "https://example.com/"
    assert list(extract_links(post)) == [
        URL("https://example.com/"),
        URL("https://nefelibata.readthedocs.io/"),
    ]



"""
Fixtures for nefelibata.
"""
# pylint: disable=invalid-name, redefined-outer-name, unused-argument
from pathlib import Path
from typing import Any, Callable, Iterator, Type

import pytest
import yaml
from freezegun import freeze_time
from pyfakefs.fake_filesystem import FakeFilesystem

from nefelibata.config import Config
from nefelibata.constants import CONFIG_FILENAME
from nefelibata.post import Post, build_post
from nefelibata.utils import get_project_root

from .fakes import CONFIG, POST_CONTENT


class MockEntryPoint:  # pylint: disable=too-few-public-methods
    """
    A class for mocking entry points.
    """

    def __init__(self, name: str, class_: Callable[..., Any]):
        self.name = name
        self.class_ = class_

    def load(self) -> Callable[..., Any]:
        """
        Return the class.
        """
        return self.class_


@pytest.fixture
def make_entry_point() -> Type[MockEntryPoint]:
    """
    Fixture for a helper function to mock entry points.
    """
    return MockEntryPoint


@pytest.fixture
def root(fs: FakeFilesystem) -> Iterator[Path]:
    """
    Create the blog root directory.
    """
    # Add the templates to the fake filesystem, so builders can load them
    # during tests. The actual path depends if we're running in development
    # mode (``src/``) or installed (``site-packages``).
    root = get_project_root()
    locations = ("src/nefelibata/templates", "site-packages/nefelibata/templates")
    for location in locations:
        try:
            fs.add_real_directory(root / location)
        except FileNotFoundError:
            pass

    root = Path("/path/to/blog")
    fs.create_dir(root)
    yield root


@pytest.fixture
def config(fs: FakeFilesystem, root: Path) -> Iterator[Config]:
    """
    Create configuration file.
    """
    fs.create_file(root / CONFIG_FILENAME, contents=yaml.dump(CONFIG))

    yield Config(**CONFIG)


@pytest.fixture
def post(fs: FakeFilesystem, root: Path, config: Config) -> Iterator[Post]:
    """
    Create a post.
    """
    post_directory = root / "posts/first"
    post_path = post_directory / "index.mkd"
    with freeze_time("2021-01-01T00:00:00Z"):
        fs.create_file(post_path, contents=POST_CONTENT)
        post = build_post(root, config, post_path)

    yield post



"""
Tests for ``nefelibata.publishers.s3``.
"""
# pylint: disable=invalid-name
from datetime import datetime, timezone
from pathlib import Path

import pytest
from botocore.exceptions import ClientError
from freezegun import freeze_time
from pyfakefs.fake_filesystem import FakeFilesystem
from pytest_mock import MockerFixture

from nefelibata.config import Config
from nefelibata.publishers.s3 import S3Publisher


@pytest.mark.asyncio
async def test_publish(
    mocker: MockerFixture,
    fs: FakeFilesystem,
    root: Path,
    config: Config,
) -> None:
    """
    Test ``publish``.
    """
    mocker.patch.object(
        S3Publisher,
        "find_modified_files",
        return_value=iter(
            [
                root / "build/generic/one/index.html",
                root / "build/generic/one/photo.jpg",
                root / "build/generic/one/blob",
            ],
        ),
    )
    boto3 = mocker.patch("nefelibata.publishers.s3.boto3")

    fs.create_file(root / "build/generic/one", contents="Hello, world!")

    publisher = S3Publisher(
        root,
        config,
        "generic",
        "KEY",
        "SECRET",
        "example.com",
    )

    with freeze_time("2021-01-01T00:00:00Z"):
        publishing = await publisher.publish()
    assert publishing.dict() == {
        "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
    }
    boto3.client.return_value.upload_file.assert_has_calls(
        [
            mocker.call(
                "/path/to/blog/build/generic/one/index.html",
                "example.com",
                "one/index.html",
                ExtraArgs={"ACL": "public-read", "ContentType": "text/html"},
            ),
            mocker.call(
                "/path/to/blog/build/generic/one/photo.jpg",
                "example.com",
                "one/photo.jpg",
                ExtraArgs={"ACL": "public-read", "ContentType": "image/jpeg"},
            ),
            mocker.call(
                "/path/to/blog/build/generic/one/blob",
                "example.com",
                "one/blob",
                ExtraArgs={"ACL": "public-read"},
            ),
        ],
    )


@pytest.mark.asyncio
async def test_publish_no_modified_files(
    mocker: MockerFixture,
    root: Path,
    config: Config,
) -> None:
    """
    Test ``publish`` when no files have been modified.
    """
    mocker.patch.object(
        S3Publisher,
        "find_modified_files",
        return_value=iter([]),
    )
    mocker.patch("nefelibata.publishers.s3.boto3")

    publisher = S3Publisher(
        root,
        config,
        "generic",
        "KEY",
        "SECRET",
        "example.com",
    )

    with freeze_time("2021-01-01T00:00:00Z"):
        publishing = await publisher.publish()
    assert publishing is None


@pytest.mark.asyncio
async def test_publish_error(
    mocker: MockerFixture,
    root: Path,
    config: Config,
) -> None:
    """
    Test ``publish`` when an error occurs.
    """
    mocker.patch.object(
        S3Publisher,
        "find_modified_files",
        return_value=iter([root / "build/generic/one/index.html"]),
    )
    boto3 = mocker.patch("nefelibata.publishers.s3.boto3")
    boto3.client.return_value.upload_file.side_effect = ClientError(
        operation_name="upload_file",
        error_response={"Error": {"Code": 42, "Message": "A wild exception appears!"}},
    )
    _logger = mocker.patch("nefelibata.publishers.s3._logger")

    publisher = S3Publisher(
        root,
        config,
        "generic",
        "KEY",
        "SECRET",
        "example.com",
    )

    with pytest.raises(ClientError) as excinfo:
        await publisher.publish()
    assert str(excinfo.value) == (
        "An error occurred (42) when calling the upload_file operation: "
        "A wild exception appears!"
    )
    _logger.exception.assert_called_with("An error occurred")



"""
Tests for ``nefelibata.console``.
"""
import asyncio
from pathlib import Path

import pytest
from pytest_mock import MockerFixture

from nefelibata import console


@pytest.mark.asyncio
async def test_main_init(mocker: MockerFixture) -> None:
    """
    Test ``main`` with the "init" action.
    """
    init = mocker.patch("nefelibata.console.init")
    init.run = mocker.AsyncMock()

    mocker.patch(
        "nefelibata.console.docopt",
        return_value={
            "--loglevel": "debug",
            "init": True,
            "new": False,
            "build": False,
            "publish": False,
            "ROOT_DIR": "/path/to/blog",
            "--force": False,
        },
    )
    await console.main()
    init.run.assert_called_with(Path("/path/to/blog"), False)


@pytest.mark.asyncio
async def test_main_new(mocker: MockerFixture) -> None:
    """
    Test ``main`` with the "new" action.
    """
    new = mocker.patch("nefelibata.console.new")
    new.run = mocker.AsyncMock()

    mocker.patch(
        "nefelibata.console.docopt",
        return_value={
            "--loglevel": "debug",
            "init": False,
            "new": True,
            "build": False,
            "publish": False,
            "ROOT_DIR": "/path/to/blog",
            "POST": "A like",
            "-t": "like",
            "--force": False,
        },
    )
    await console.main()
    new.run.assert_called_with(Path("/path/to/blog"), "A like", "like")


@pytest.mark.asyncio
async def test_main_build(mocker: MockerFixture) -> None:
    """
    Test ``main`` with the "build" action.
    """
    build = mocker.patch("nefelibata.console.build")
    build.run = mocker.AsyncMock()

    mocker.patch(
        "nefelibata.console.docopt",
        return_value={
            "--loglevel": "debug",
            "init": False,
            "new": False,
            "build": True,
            "publish": False,
            "ROOT_DIR": "/path/to/blog",
            "--force": False,
        },
    )
    await console.main()
    build.run.assert_called_with(Path("/path/to/blog"), False)

    mocker.patch(
        "nefelibata.console.docopt",
        return_value={
            "--loglevel": "debug",
            "init": False,
            "new": False,
            "build": True,
            "publish": False,
            "ROOT_DIR": "/path/to/blog",
            "--force": True,
        },
    )
    await console.main()
    build.run.assert_called_with(Path("/path/to/blog"), True)

    mocker.patch(
        "nefelibata.console.docopt",
        return_value={
            "--loglevel": "debug",
            "init": False,
            "new": False,
            "build": True,
            "publish": False,
            "ROOT_DIR": None,
            "--force": True,
        },
    )
    mocker.patch(
        "nefelibata.console.find_directory",
        return_value=Path("/path/to/blog"),
    )
    await console.main()
    build.run.assert_called_with(Path("/path/to/blog"), True)


@pytest.mark.asyncio
async def test_main_publish(mocker: MockerFixture) -> None:
    """
    Test ``main`` with the "publish" action.
    """
    publish = mocker.patch("nefelibata.console.publish")
    publish.run = mocker.AsyncMock()

    mocker.patch(
        "nefelibata.console.docopt",
        return_value={
            "--loglevel": "debug",
            "init": False,
            "new": False,
            "build": False,
            "publish": True,
            "ROOT_DIR": "/path/to/blog",
            "POST": "A like",
            "-t": "like",
            "--force": False,
        },
    )
    await console.main()
    publish.run.assert_called_with(Path("/path/to/blog"), False)


@pytest.mark.asyncio
async def test_main_no_action(mocker: MockerFixture) -> None:
    """
    Test ``main`` without any actions -- should not happen.
    """
    mocker.patch(
        "nefelibata.console.docopt",
        return_value={
            "--loglevel": "debug",
            "init": False,
            "new": False,
            "build": False,
            "publish": False,
            "ROOT_DIR": "/path/to/blog",
            "--force": False,
        },
    )
    await console.main()


@pytest.mark.asyncio
async def test_main_canceled(mocker: MockerFixture) -> None:
    """
    Test canceling the ``main`` coroutine.
    """
    build = mocker.patch("nefelibata.console.build")
    build.run = mocker.AsyncMock(side_effect=asyncio.CancelledError("Canceled"))
    _logger = mocker.patch("nefelibata.console._logger")

    mocker.patch(
        "nefelibata.console.docopt",
        return_value={
            "--loglevel": "debug",
            "init": False,
            "new": False,
            "build": True,
            "publish": False,
            "ROOT_DIR": "/path/to/blog",
            "--force": False,
        },
    )
    await console.main()

    _logger.info.assert_called_with("Canceled")


def test_run(mocker: MockerFixture) -> None:
    """
    Test ``run``.
    """
    main = mocker.AsyncMock()
    mocker.patch("nefelibata.console.main", main)

    console.run()

    main.assert_called()


def test_interrupt(mocker: MockerFixture) -> None:
    """
    Test that ``CTRL-C`` stops the CLI.
    """
    main = mocker.AsyncMock(side_effect=KeyboardInterrupt())
    mocker.patch("nefelibata.console.main", main)
    _logger = mocker.patch("nefelibata.console._logger")

    console.run()

    _logger.info.assert_called_with("Stopping Nefelibata")



"""
Tests for ``nefelibata.announcers.mastodon``.
"""
# pylint: disable=invalid-name

from datetime import datetime, timezone
from pathlib import Path

import pytest
from mastodon import MastodonNotFoundError
from pytest_mock import MockerFixture

from nefelibata.announcers.base import Author, Interaction
from nefelibata.announcers.mastodon import MastodonAnnouncer
from nefelibata.builders.base import Builder
from nefelibata.config import Config
from nefelibata.enclosure import Enclosure
from nefelibata.post import Post


@pytest.mark.asyncio
async def test_announcer_announce(
    mocker: MockerFixture,
    root: Path,
    config: Config,
    post: Post,
) -> None:
    """
    Test announcing posts.
    """
    gemini_builder = Builder(root, config, "gemini://example.com/", "gemini")
    gemini_builder.extension = ".gmi"
    html_builder = Builder(root, config, "https://example.com/", "www")
    html_builder.extension = ".html"

    Mastodon = mocker.patch("nefelibata.announcers.mastodon.Mastodon")
    Mastodon.return_value.status_post.return_value.url = "https://example.com/1"
    Mastodon.return_value.status_post.return_value.created_at = datetime(
        2021,
        1,
        1,
        tzinfo=timezone.utc,
    )

    announcer = MastodonAnnouncer(
        root,
        config,
        builders=[gemini_builder, html_builder],
        access_token="access_token",
        base_url="https://example.com/",
    )

    enclosure_path = post.path.parent / "picture.jpg"
    enclosure_path.touch()
    post.enclosures.append(
        Enclosure(
            path=enclosure_path,
            description="A photo",
            type="image/jpeg",
            length=666,
            href="first/picture.jpg",
        ),
    )

    announcement = await announcer.announce_post(post)
    assert announcement.dict() == {
        "url": "https://example.com/1",
        "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
        "grace_seconds": 0,
    }

    Mastodon.return_value.media_post.assert_called_with(
        str(enclosure_path),
        "image/jpeg",
        "A photo",
    )
    Mastodon.return_value.status_post.assert_called_with(
        status="""Hello, world!

gemini://example.com/first/index.gmi
https://example.com/first/index.html""",
        visibility="public",
        media_ids=[Mastodon.return_value.media_post.return_value],
        language="en",
        idempotency_key=str(post.path),
    )


@pytest.mark.asyncio
async def test_announcer_announce_multiple_enclosures(
    mocker: MockerFixture,
    root: Path,
    config: Config,
    post: Post,
) -> None:
    """
    Test announcing posts with 4+ enclosures.
    """
    gemini_builder = Builder(root, config, "gemini://example.com/", "gemini")
    gemini_builder.extension = ".gmi"

    Mastodon = mocker.patch("nefelibata.announcers.mastodon.Mastodon")
    Mastodon.return_value.status_post.return_value.url = "https://example.com/1"
    Mastodon.return_value.status_post.return_value.created_at = datetime(
        2021,
        1,
        1,
        tzinfo=timezone.utc,
    )
    _logger = mocker.patch("nefelibata.announcers.mastodon._logger")

    announcer = MastodonAnnouncer(
        root,
        config,
        builders=[gemini_builder],
        access_token="access_token",
        base_url="https://example.com/",
    )

    for i in range(5):
        enclosure_path = post.path.parent / f"picture{i}.jpg"
        enclosure_path.touch()
        post.enclosures.append(
            Enclosure(
                path=enclosure_path,
                description=f"Photo #{i}",
                type="image/jpeg",
                length=666,
                href=f"first/picture{i}.jpg",
            ),
        )

    await announcer.announce_post(post)

    _logger.warning.assert_called_with(
        "Found more than %d media enclosures in post %s. Only the first "
        "%d will be uploaded.",
        4,
        post.path,
        4,
    )


@pytest.mark.asyncio
async def test_announcer_collect(
    mocker: MockerFixture,
    root: Path,
    config: Config,
    post: Post,
) -> None:
    """
    Test collecting interactions on a given post.
    """
    Mastodon = mocker.patch("nefelibata.announcers.mastodon.Mastodon")
    Mastodon.return_value.status_context.return_value = {
        "descendants": [
            {
                "uri": "https://example.com/2",
                "id": 2,
                "url": "https://example.com/2",
                "content": "This is a reply",
                "created_at": datetime(2021, 1, 1),
                "account": {
                    "display_name": "Alice Doe",
                    "url": "https://alice.example.com/",
                    "avatar": "https://alice.example.com/profile.gif",
                    "note": "Best friends with Bob and Charlie",
                },
                "in_reply_to_id": 1,
            },
        ],
    }

    announcer = MastodonAnnouncer(
        root,
        config,
        builders=[],
        access_token="access_token",
        base_url="https://example.com/",
    )

    post.metadata["announcements"] = {
        "mastodon": {
            "url": "https://example.com/1",
        },
        "something": {
            "url": "https://foo.example.com/",
        },
    }

    interactions = await announcer.collect_post(post)
    assert interactions == {
        "https://example.com/2": Interaction(
            id="https://example.com/2",
            name="https://example.com/2",
            summary=None,
            content="This is a reply",
            published=datetime(2021, 1, 1, 0, 0),
            updated=None,
            author=Author(
                name="Alice Doe",
                url="https://alice.example.com/",
                avatar="https://alice.example.com/profile.gif",
                note="Best friends with Bob and Charlie",
            ),
            url="https://example.com/2",
            in_reply_to="https://example.com/1",
            type="reply",
        ),
    }


@pytest.mark.asyncio
async def test_announcer_collect_not_found(
    mocker: MockerFixture,
    root: Path,
    config: Config,
    post: Post,
) -> None:
    """
    Test collecting interactions when post has been deleted.
    """
    Mastodon = mocker.patch("nefelibata.announcers.mastodon.Mastodon")
    Mastodon.return_value.status_context.side_effect = MastodonNotFoundError(
        "Not found",
    )
    _logger = mocker.patch("nefelibata.announcers.mastodon._logger")

    announcer = MastodonAnnouncer(
        root,
        config,
        builders=[],
        access_token="access_token",
        base_url="https://example.com/",
    )

    post.metadata["announcements"] = {
        "mastodon": {
            "url": "https://example.com/1",
        },
    }

    interactions = await announcer.collect_post(post)
    assert interactions == {}

    _logger.warning.assert_called_with("Toot %s not found", "https://example.com/1")



"""
Assistant for mirrorring images locally.
"""

import asyncio
import hashlib
import logging
import mimetypes
from collections import defaultdict
from io import BytesIO
from pathlib import Path
from typing import Dict, Iterator, Tuple

import marko
import piexif
from aiohttp import ClientSession
from PIL import Image

from nefelibata.announcers.base import Scope
from nefelibata.assistants.base import Assistant
from nefelibata.post import Post

_logger = logging.getLogger(__name__)

CHUNK_SIZE = 2048


def extract_images(content: str) -> Iterator[Tuple[str, str]]:
    """
    Extract all images from a Markdown document.
    """
    tree = marko.parse(content)
    queue = [tree]
    while queue:
        element = queue.pop()

        if isinstance(element, marko.inline.Image):
            yield element.dest, element.children[0].children
        elif hasattr(element, "children"):
            queue.extend(element.children)


def is_local(url: str) -> bool:
    """
    Return true if the image is local.

    For now we return true if the URL is relative.
    """
    return "://" not in url


async def get_resource_extension(session: ClientSession, url: str) -> str:
    """
    Return the extension of a remote image.
    """
    async with session.head(url) as response:
        content_type = response.headers["content-type"]

    extension = mimetypes.guess_extension(content_type)
    return extension or ""


def get_filename(url: str, extension: str) -> str:
    """
    Compute the filename for a given resource.
    """
    md5 = hashlib.md5()
    md5.update(url.encode("utf-8"))
    return f"{md5.hexdigest()}{extension}"


def add_exif(buf: BytesIO, url: str, title: str) -> BytesIO:
    """
    Store the URL in the EXIF data.
    """
    buf.seek(0)
    image = Image.open(buf)

    exif = (
        piexif.load(image.info["exif"]) if "exif" in image.info else defaultdict(dict)
    )
    exif["0th"][piexif.ImageIFD.Copyright] = url
    exif["0th"][piexif.ImageIFD.ImageDescription] = title

    _logger.info("Adding original URL and title to the EXIF data")
    buf = BytesIO()
    image.save(buf, "jpeg", exif=piexif.dump(exif))

    return buf


async def download_image(  # pylint: disable=too-many-arguments
    session: ClientSession,
    url: str,
    title: str,
    post: Post,
    directory: Path,
    replacements: Dict[str, str],
) -> None:
    """
    Download an image.
    """
    extension = await get_resource_extension(session, url)
    filename = get_filename(url, extension)
    target = directory / filename

    if target.exists():
        _logger.debug("Image already mirrored")
        return

    replacements[url] = str(target.relative_to(post.path.parent))

    _logger.info("Downloading image from %s", url)
    buf = BytesIO()
    async with session.get(url) as response:
        async for chunk in response.content.iter_chunked(  # pragma: no cover
            CHUNK_SIZE,
        ):
            buf.write(chunk)

    if extension in {".jpeg", ".jpg"}:
        buf = add_exif(buf, url, title)

    with open(target, "wb") as output:
        output.write(buf.getvalue())


class MirrorImagesAssistant(Assistant):
    """
    Assistant for mirrorring images locally.
    """

    name = "mirror_images"
    scopes = [Scope.POST]

    async def process_post(self, post: Post, force: bool = False) -> None:
        # create directory for images
        mirror = post.path.parent / "img"
        if not mirror.exists():
            mirror.mkdir()

        replacements: Dict[str, str] = {}
        tasks = []
        async with ClientSession() as session:
            for url, title in extract_images(post.content):
                if is_local(url):
                    continue

                task = asyncio.create_task(
                    download_image(session, url, title, post, mirror, replacements),
                )
                tasks.append(task)

            await asyncio.gather(*tasks)

        if not replacements:
            return

        _logger.info("Updating images in file %s", post.path)
        async with post._lock:  # pylint: disable=protected-access
            with open(post.path, encoding="utf-8") as input_:
                content = input_.read()

            for original, new in replacements.items():
                content = content.replace(original, new)

            with open(post.path, "w", encoding="utf-8") as output:
                output.write(content)



"""
Tests for ``nefelibata.publishers.base``.
"""
# pylint: disable=invalid-name
from datetime import datetime, timezone
from pathlib import Path
from typing import Optional, Type

from freezegun import freeze_time
from pyfakefs.fake_filesystem import FakeFilesystem
from pytest_mock import MockerFixture

from nefelibata.config import BuilderModel, Config, PublisherModel
from nefelibata.publishers.base import Publisher, get_publishers

from ..conftest import MockEntryPoint


def test_get_publishers(
    mocker: MockerFixture,
    make_entry_point: Type[MockEntryPoint],
    root: Path,
    config: Config,
) -> None:
    """
    Test ``get_publishers``.
    """

    class DummyPublisher(Publisher):
        """
        A dummy publisher.
        """

        async def publish(
            self,
            since: Optional[datetime] = None,
            force: bool = False,
        ) -> None:
            pass

    entry_points = [make_entry_point("dummy_publisher", DummyPublisher)]
    mocker.patch(
        "nefelibata.publishers.base.iter_entry_points",
        return_value=entry_points,
    )

    config.publishers = {"dummy": PublisherModel(plugin="dummy_publisher")}
    config.builders = {
        "site_builder": BuilderModel(
            **{
                "plugin": "site_builder",
                "announce-on": [],
                "publish-to": ["dummy"],
                "home": "https://example.com/",
                "path": "site",
            }
        ),
    }
    publishers = get_publishers(root, config)
    assert len(publishers) == 1
    assert isinstance(publishers["site_builder => dummy"], DummyPublisher)


def test_find_modified_files(
    fs: FakeFilesystem,
    root: Path,
    config: Config,
) -> None:
    """
    Test ``find_modified_files``.
    """
    publisher = Publisher(root, config, "generic")

    fs.create_dir(root / "build/generic")
    fs.create_dir(root / "build/generic/subdir")
    with freeze_time("2021-01-01T00:00:00Z"):
        (root / "build/generic/one").touch()
    with freeze_time("2021-01-02T00:00:00Z"):
        (root / "build/generic/subdir/two").touch()

    since = datetime(2021, 1, 1, 12, 0, 0, tzinfo=timezone.utc)
    modified_files = list(publisher.find_modified_files(force=False, since=since))
    assert modified_files == [Path("/path/to/blog/build/generic/subdir/two")]

    since = datetime(2021, 1, 2, 12, 0, 0, tzinfo=timezone.utc)
    modified_files = list(publisher.find_modified_files(force=False, since=since))
    assert modified_files == []

    since = datetime(2021, 1, 2, 12, 0, 0, tzinfo=timezone.utc)
    modified_files = list(publisher.find_modified_files(force=True, since=since))
    assert modified_files == [
        Path("/path/to/blog/build/generic/one"),
        Path("/path/to/blog/build/generic/subdir/two"),
    ]



"""
Assistant for generating a playlist file.
"""

from pathlib import Path
from typing import Any

from nefelibata.announcers.base import Scope
from nefelibata.assistants.base import Assistant
from nefelibata.config import Config
from nefelibata.post import Post


class PlaylistAssistant(Assistant):
    """
    Assistant for generating a playlist file.
    """

    name = "playlist"
    scopes = [Scope.POST]

    def __init__(self, root: Path, config: Config, base_url: str, **kwargs: Any):
        super().__init__(root, config, **kwargs)

        self.base_url = base_url.rstrip("/")

    async def process_post(self, post: Post, force: bool = False) -> None:
        # create a playlist for each builder
        path = post.path.parent / "index.pls"
        if path.exists() and not force:
            return

        valid_enclosures = [
            enclosure for enclosure in post.enclosures if enclosure.type == "audio/mpeg"
        ]
        if not valid_enclosures:
            return

        with open(path, "w", encoding="utf-8") as output:
            output.write(
                f"""[playlist]

NumberOfEntries={len(valid_enclosures)}
Version=2

""",
            )
            for i, enclosure in enumerate(valid_enclosures):
                output.write(f"File{i+1}={self.base_url}/{enclosure.href}\n")
                output.write(f"Title{i+1}={enclosure.description}\n")
                output.write(f"Length{i+1}={enclosure.duration}\n\n")



"""
Tests for ``nefelibata.enclosure``.
"""
# pylint: disable=invalid-name

from pathlib import Path

from pyfakefs.fake_filesystem import FakeFilesystem
from pytest_mock import MockerFixture

from nefelibata.enclosure import get_enclosures, get_pretty_duration


def test_get_pretty_duration() -> None:
    """
    Test ``get_pretty_duration``.
    """
    assert get_pretty_duration(0) == "0s"
    assert get_pretty_duration(1) == "1s"
    assert get_pretty_duration(60) == "1m"
    assert get_pretty_duration(61) == "1m1s"
    assert get_pretty_duration(120) == "2m"
    assert get_pretty_duration(121) == "2m1s"
    assert get_pretty_duration(3600) == "1h"
    assert get_pretty_duration(3601) == "1h0m1s"


def test_get_enclosures(mocker: MockerFixture, fs: FakeFilesystem, root: Path) -> None:
    """
    Test ``get_enclosures``.
    """
    MP3 = mocker.patch("nefelibata.enclosure.MP3")
    metadata = {
        "TIT2": "A title",
        "TPE1": "An artist",
        "TALB": "An album",
        "TDRC": 2021,
        "TRCK": 1,
    }
    MP3.return_value.get.side_effect = metadata.get
    MP3.return_value.info.length = 123.0

    piexif = mocker.patch("nefelibata.enclosure.piexif")
    piexif.ImageIFD.ImageDescription = 270
    piexif.load.return_value = {"0th": {270: "This is a nice photo"}}

    fs.create_dir(root / "posts/first")
    path = root / "posts/first/index.mkd"

    # create supported files
    (root / "posts/first/song.mp3").touch()
    (root / "posts/first/photo.jpg").touch()
    (root / "posts/first/logo.png").touch()

    # create non-supported file
    (root / "posts/first/test.txt").touch()

    enclosures = get_enclosures(root, path.parent)
    assert len(enclosures) == 3

    assert enclosures[0].dict() == {
        "path": Path("/path/to/blog/posts/first/song.mp3"),
        "description": '"A title" (2m3s) by An artist (An album, 2021)',
        "type": "audio/mpeg",
        "length": 0,
        "href": "first/song.mp3",
        "title": "A title",
        "artist": "An artist",
        "album": "An album",
        "year": 2021,
        "duration": 123.0,
        "track": 0,
    }

    assert enclosures[1].dict() == {
        "description": "This is a nice photo",
        "href": "first/photo.jpg",
        "length": 0,
        "path": Path("/path/to/blog/posts/first/photo.jpg"),
        "type": "image/jpeg",
    }

    assert enclosures[2].dict() == {
        "description": "Image logo.png",
        "href": "first/logo.png",
        "length": 0,
        "path": Path("/path/to/blog/posts/first/logo.png"),
        "type": "image/png",
    }



"""
An Antenna (gemini://warmedal.se/~antenna/) announcer.
"""

import logging

from nefelibata.announcers.gemlog import GemlogAnnouncer

_logger = logging.getLogger(__name__)


class AntennaAnnouncer(GemlogAnnouncer):

    """
    An Antenna (gemini://warmedal.se/~antenna/) announcer.

    This requires the Gemini builder, and assumes that the Gemlog feed is
    at ``{builder.home}/feed{builder.extension}``.
    """

    name = "Antenna"
    url = "gemini://warmedal.se/~antenna/"
    submit_url = "gemini://warmedal.se/~antenna/submit?"
    logger = _logger



"""
A Mastodon announcer.
"""

import logging
from pathlib import Path
from typing import Any, Dict, List, Optional

from mastodon import Mastodon, MastodonNotFoundError

from nefelibata.announcers.base import (
    Announcement,
    Announcer,
    Author,
    Interaction,
    Scope,
)
from nefelibata.builders.base import Builder
from nefelibata.config import Config
from nefelibata.post import Post

_logger = logging.getLogger(__name__)

VALID_MIMETYPES = {"audio/mpeg", "image/jpeg"}
MAX_NUMBER_OF_ENCLOSURES = 4


class MastodonAnnouncer(Announcer):

    """
    A Mastodon announcer.
    """

    scopes = [Scope.POST]

    def __init__(  # pylint: disable=too-many-arguments
        self,
        root: Path,
        config: Config,
        builders: List[Builder],
        access_token: str,
        base_url: str,
        **kwargs: Any,
    ):
        super().__init__(root, config, builders, **kwargs)

        self.base_url = base_url

        self.client = Mastodon(
            access_token=access_token,
            api_base_url=base_url,
        )

    async def announce_post(self, post: Post) -> Optional[Announcement]:
        valid_enclosures = [
            enclosure
            for enclosure in post.enclosures
            if enclosure.type in VALID_MIMETYPES
        ]
        if len(valid_enclosures) > MAX_NUMBER_OF_ENCLOSURES:
            _logger.warning(
                "Found more than %d media enclosures in post %s. Only the first "
                "%d will be uploaded.",
                MAX_NUMBER_OF_ENCLOSURES,
                post.path,
                MAX_NUMBER_OF_ENCLOSURES,
            )

        media_ids = []
        for enclosure in valid_enclosures[:MAX_NUMBER_OF_ENCLOSURES]:
            _logger.info("Uploading post enclosure %s", enclosure.path)

            media_dict = self.client.media_post(
                str(enclosure.path),
                enclosure.type,
                enclosure.description,
            )
            media_ids.append(media_dict)

        language = post.metadata.get("language") or self.config.language
        summary = post.metadata.get("summary") or post.title
        urls = "\n".join(
            builder.absolute_url(post).human_repr() for builder in self.builders
        )
        status = f"{summary}\n\n{urls}"

        _logger.info("Announcing post %s on Mastodon", post.path)
        toot_dict = self.client.status_post(
            status=status,
            visibility="public",
            media_ids=media_ids,
            language=language,
            idempotency_key=str(post.path),
        )

        return Announcement(
            url=toot_dict.url,
            timestamp=toot_dict.created_at,
        )

    async def collect_post(self, post: Post) -> Dict[str, Interaction]:
        interactions: Dict[str, Interaction] = {}

        announcements = post.metadata.get("announcements", {})
        for announcement in announcements.values():
            if not announcement["url"].startswith(self.base_url):
                continue

            url = announcement["url"]
            id_ = int(url.rstrip("/").rsplit("/", 1)[1])
            try:
                context = self.client.status_context(id_)
            except MastodonNotFoundError:
                _logger.warning("Toot %s not found", url)
                continue

            # map from instance ID to URL
            id_map = {id_: url}

            for descendant in context["descendants"]:
                id_ = descendant["url"]
                id_map[descendant["id"]] = id_
                interactions[str(id_)] = Interaction(
                    id=descendant["uri"],
                    name=descendant["url"],
                    summary=None,
                    content=descendant["content"],
                    published=descendant["created_at"],
                    updated=None,
                    author=Author(
                        name=descendant["account"]["display_name"],
                        url=descendant["account"]["url"],
                        avatar=descendant["account"]["avatar"],
                        note=descendant["account"]["note"],
                    ),
                    url=descendant["url"],
                    in_reply_to=id_map[descendant["in_reply_to_id"]],
                    type="reply",
                )

        return interactions



"""
Assistant for computing reading time.
"""

import math
from typing import Any, Dict

from nefelibata.announcers.base import Scope
from nefelibata.assistants.base import Assistant
from nefelibata.post import Post


class ReadingTimeAssistant(Assistant):
    """
    An assistant for computing post reading time.

    This is based on a formula supposedly used by Medium:

        https://www.quora.com/How-does-Medium-determine-an-article%E2%80%99s-estimated-read-time/answer/Alan-Hamlett
    """

    name = "reading_time"
    scopes = [Scope.POST]

    async def get_post_metadata(self, post: Post) -> Dict[str, Any]:
        num_images = len(
            [
                enclosure
                for enclosure in post.enclosures
                if enclosure.type.startswith("image")
            ],
        )
        num_words = len(post.content.split(" "))
        image_weight = max(13 - num_images, 3)
        seconds = num_words / 265 * 60 + image_weight * num_images
        minutes = math.ceil(seconds / 60)

        return {
            "words": num_words,
            "images": num_images,
            "total_seconds": seconds,
            "total_minutes": minutes,
        }



"""
Tests for ``nefelibata.utils``.
"""
# pylint: disable=invalid-name

import logging
from pathlib import Path

import pytest
import yaml
from pydantic import BaseModel
from pyfakefs.fake_filesystem import FakeFilesystem
from pytest_mock import MockerFixture
from yarl import URL

from nefelibata.announcers.base import Interaction
from nefelibata.config import Config
from nefelibata.utils import (
    archive_urls,
    dict_merge,
    find_directory,
    get_config,
    load_extra_metadata,
    load_yaml,
    setup_logging,
    update_yaml,
)


def test_setup_logging() -> None:
    """
    Test ``setup_logging``.
    """
    setup_logging("debug")
    assert logging.root.level == logging.DEBUG

    with pytest.raises(ValueError) as excinfo:
        setup_logging("invalid")
    assert str(excinfo.value) == "Invalid log level: invalid"


def test_find_directory(fs: FakeFilesystem) -> None:
    """
    Test ``find_directory``.
    """
    fs.create_dir("/path/to/blog/posts/first/css")
    fs.create_file("/path/to/blog/nefelibata.yaml")

    path = find_directory(Path("/path/to/blog/posts/first/css"))
    assert path == Path("/path/to/blog")

    with pytest.raises(SystemExit) as excinfo:
        find_directory(Path("/path/to"))
    assert str(excinfo.value) == "No configuration found!"


def test_get_config(root: Path, config: Config) -> None:
    """
    Test ``get_config``.
    """
    config = get_config(root)
    assert config.dict() == {
        "author": {
            "email": "roberto@dealmeida.net",
            "name": "Beto Dealmeida",
            "note": "Este, sou eu",
            "url": "https://taoetc.org/",
        },
        "title": "道&c.",
        "subtitle": "Musings about the path and other things",
        "language": "en",
        "categories": {
            "stem": {
                "description": "Science, technology, engineering, & math",
                "label": "STEM",
                "tags": ["blog", "programming"],
            },
        },
        "announcers": {"announcer": {"plugin": "announcer"}},
        "assistants": {"assistant": {"plugin": "assistant"}},
        "builders": {
            "builder": {
                "announce_on": ["announcer"],
                "home": "https://example.com/",
                "path": "generic",
                "plugin": "builder",
                "publish_to": ["publisher"],
            },
        },
        "publishers": {"publisher": {"plugin": "publisher"}},
        "social": [{"title": "My page", "url": "https://example.com/user"}],
        "templates": {"short": []},
    }

    with pytest.raises(SystemExit) as excinfo:
        get_config(Path("/path/to"))
    assert str(excinfo.value) == "No configuration found!"


def test_load_yaml(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test ``load_yaml``.
    """
    assert load_yaml(Path("/path/to/blog/missing.yaml"), BaseModel) == {}

    fs.create_file(
        "/path/to/blog/existing.yaml",
        contents="""
reply,gemini://ew.srht.site/en/2021/20210915-re-changing-old-code-is-risky.gmi:
  id: reply,gemini://ew.srht.site/en/2021/20210915-re-changing-old-code-is-risky.gmi
  name: '2021-09-15 ~ew''s FlightLog: Re: Changing old code is risky'
  timestamp: null
  type: reply
  url: gemini://ew.srht.site/en/2021/20210915-re-changing-old-code-is-risky.gmi
    """,
    )
    assert load_yaml(Path("/path/to/blog/existing.yaml"), Interaction) == {
        (
            "reply,gemini://ew.srht.site/en/2021/"
            "20210915-re-changing-old-code-is-risky.gmi"
        ): Interaction(
            id="reply,gemini://ew.srht.site/en/2021/20210915-re-changing-old-code-is-risky.gmi",
            name="2021-09-15 ~ew's FlightLog: Re: Changing old code is risky",
            url="gemini://ew.srht.site/en/2021/20210915-re-changing-old-code-is-risky.gmi",
            type="reply",
            timestamp=None,
        ),
    }

    path = Path("/path/to/blog/invalid.yaml")
    fs.create_file(path, contents="[1,2,3")
    _logger = mocker.patch("nefelibata.utils._logger")
    with pytest.raises(yaml.parser.ParserError):
        load_yaml(path, BaseModel)
    assert _logger.warning.called_with("Invalid YAML file: %s", path)


def test_dict_merge() -> None:
    """
    Test ``dict_merge``.
    """
    original = {"a": {"b": "c"}}
    update = {"d": {"e": "f"}, "a": {"g": "h"}}
    dict_merge(original, update)
    assert original == {"a": {"b": "c", "g": "h"}, "d": {"e": "f"}}


def test_load_extra_metadata(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test ``load_extra_metadata``.
    """
    _logger = mocker.patch("nefelibata.utils._logger")

    fs.create_file("/path/to/blog/test.yaml", contents=yaml.dump(dict(a=42)))
    fs.create_file("/path/to/blog/broken.yaml", contents="{[")

    metadata = load_extra_metadata(Path("/path/to/blog"))
    assert metadata == {"test": {"a": 42}}

    _logger.warning.assert_called_with(
        "Invalid file: %s",
        Path("/path/to/blog/broken.yaml"),
    )


def test_update_yaml(fs: FakeFilesystem) -> None:
    """
    Test ``update_yaml``.
    """
    path = Path("/path/to/blog/test.yaml")
    fs.create_dir(path.parent)

    # create new file
    with update_yaml(path) as config:
        config["foo"] = "bar"

    # check contents
    with update_yaml(path) as config:
        assert config["foo"] == "bar"

    # test error
    with open(path, "w", encoding="utf-8") as output:
        output.write("{{ test }}")
    with pytest.raises(yaml.constructor.ConstructorError) as excinfo:
        with update_yaml(path) as config:
            pass
    assert (
        str(excinfo.value)
        == """while constructing a mapping
  in "<unicode string>", line 1, column 1:
    {{ test }}
    ^
found unhashable key
  in "<unicode string>", line 1, column 2:
    {{ test }}
     ^"""
    )


@pytest.mark.asyncio
async def test_archive_urls(mocker: MockerFixture) -> None:
    """
    Test ``archive_urls``.
    """
    links = {
        "original": {
            "rel": "original",
            "url": URL("https://nefelibata.readthedocs.io/"),
        },
        "timemap": {
            "rel": "timemap",
            "type": "application/link-format",
            "url": URL(
                "https://web.archive.org/web/timemap/link/https://nefelibata.readthedocs.io/",
            ),
        },
        "timegate": {
            "rel": "timegate",
            "url": URL(
                "https://web.archive.org/web/https://nefelibata.readthedocs.io/",
            ),
        },
        "first memento": {
            "rel": "first memento",
            "datetime": "Sun, 05 Aug 2018 06:22:28 GMT",
            "url": URL(
                "https://web.archive.org/web/20180805062228/http://nefelibata.readthedocs.io/",
            ),
        },
        "prev memento": {
            "rel": "prev memento",
            "datetime": "Tue, 29 Jun 2021 01:15:06 GMT",
            "url": URL(
                "https://web.archive.org/web/20210629011506/https://nefelibata.readthedocs.io/",
            ),
        },
        "memento": {
            "rel": "memento",
            "datetime": "Sun, 03 Oct 2021 15:46:02 GMT",
            "url": URL(
                "https://web.archive.org/web/20211003154602/https://nefelibata.readthedocs.io/",
            ),
        },
        "last memento": {
            "rel": "last memento",
            "datetime": "Sun, 03 Oct 2021 15:46:02 GMT",
            "url": URL(
                "https://web.archive.org/web/20211003154602/https://nefelibata.readthedocs.io/",
            ),
        },
    }
    get = mocker.patch("nefelibata.utils.ClientSession.get")
    get.return_value.__aenter__.return_value.links = links

    saved_urls = await archive_urls(
        [
            URL("https://nefelibata.readthedocs.io/"),
            URL("gemini://taoetc.org/"),
        ],
    )
    assert saved_urls == {
        URL("https://nefelibata.readthedocs.io/"): URL(
            "https://web.archive.org/web/20211003154602/https://nefelibata.readthedocs.io/",
        ),
    }


@pytest.mark.asyncio
async def test_sleep(
    mocker: MockerFixture,
) -> None:
    """
    Test that we ``sleep`` between URLs.
    """
    get = mocker.patch("nefelibata.utils.ClientSession.get")
    get.return_value.__aenter__.return_value.links = {}

    sleep = mocker.patch("nefelibata.utils.asyncio.sleep")

    await archive_urls(
        [URL("https://example.com/foo"), URL("https://example.com/bar")],
    )

    sleep.assert_called_with(12.0)



"""
Utility functions.
"""
import asyncio
import logging
from contextlib import contextmanager
from datetime import timedelta
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Set, Type

import yaml
from aiohttp import ClientSession
from pydantic import BaseModel
from rich.logging import RichHandler
from yarl import URL

from nefelibata.config import Config
from nefelibata.constants import CONFIG_FILENAME

_logger = logging.getLogger(__name__)


def setup_logging(loglevel: str) -> None:
    """
    Setup basic logging.
    """
    level = getattr(logging, loglevel.upper(), None)
    if not isinstance(level, int):
        raise ValueError(f"Invalid log level: {loglevel}")

    logformat = "[%(asctime)s] %(levelname)s: %(name)s: %(message)s"
    logging.basicConfig(
        level=level,
        format=logformat,
        datefmt="[%X]",
        handlers=[RichHandler()],
        force=True,
    )


def find_directory(cwd: Path) -> Path:
    """
    Find root of blog, starting from `cwd`.

    The function will traverse up trying to find a configuration file.
    """
    while not (cwd / CONFIG_FILENAME).exists():
        if cwd == cwd.parent:
            raise SystemExit("No configuration found!")
        cwd = cwd.parent

    return cwd


def get_config(root: Path) -> Config:
    """
    Return the configuration for a blog.
    """
    path = root / CONFIG_FILENAME
    if not path.exists():
        raise SystemExit("No configuration found!")

    with open(path, encoding="utf-8") as input_:
        config = Config(**yaml.full_load(input_))

    return config


def get_project_root() -> Path:
    """
    Return the project root.

    This is used for unit tests.
    """
    return Path(__file__).parent.parent.parent


def split_header(header: Optional[str]) -> Set[str]:
    """
    Split a comma separated list from the post header.
    """
    if header is None:
        return set()

    return {value.strip() for value in header.split(",") if value.strip()}


def load_yaml(path: Path, class_: Type[BaseModel]) -> Dict[str, BaseModel]:
    """
    Load a YAML file into a model.
    """
    if not path.exists():
        return {}

    with open(path, encoding="utf-8") as input_:
        try:
            content = yaml.load(input_, Loader=yaml.SafeLoader)
        except (AttributeError, yaml.parser.ParserError) as ex:
            _logger.warning("Invalid YAML file: %s", path)
            raise ex

    return {name: class_(**parameters) for name, parameters in content.items()}


@contextmanager
def update_yaml(path: Path) -> Iterator[Dict[Any, Any]]:
    """
    Open or create a YAML file, and save back if modified.
    """
    if path.exists():
        with open(path, encoding="utf-8") as input_:
            original = input_.read()
            try:
                content = yaml.load(original, Loader=yaml.SafeLoader)
            except (
                AttributeError,
                yaml.parser.ParserError,
                yaml.constructor.ConstructorError,
            ) as ex:
                _logger.warning("Invalid YAML file: %s", path)
                raise ex
    else:
        original = ""
        content = {}

    try:
        yield content
    finally:
        current = yaml.dump(content)
        if current != original:
            with open(path, "w", encoding="utf-8") as output:
                output.write(current)


def dict_merge(original, update):
    """
    Recursive ``dict.update``.
    """
    for key in update:
        if (
            key in original
            and isinstance(original[key], dict)
            and isinstance(update[key], dict)
        ):
            dict_merge(original[key], update[key])
        else:
            original[key] = update[key]


def load_extra_metadata(post_directory: Path) -> Dict[str, Any]:
    """
    Load all YAML files with extra metadata for a given path.
    """
    extra_metadata = {}
    for file_path in post_directory.glob("*.yaml"):
        with open(file_path, encoding="utf-8") as input_:
            try:
                content = yaml.load(input_, Loader=yaml.SafeLoader)
            except (AttributeError, yaml.parser.ParserError):
                _logger.warning("Invalid file: %s", file_path)
                continue
        extra_metadata[file_path.stem] = content

    return extra_metadata


async def archive_urls(
    urls: List[URL],
    sleep: timedelta = timedelta(seconds=12),
) -> Dict[URL, URL]:
    """
    Save a list of URLs in https://archive.org/.
    """
    # API is restricted to 5 requests per minute, see
    # https://rationalwiki.org/wiki/Internet_Archive#Restrictions
    lock = asyncio.Lock()

    saved_urls = {}

    async with ClientSession() as session:
        for i, url in enumerate(urls):
            if url.scheme not in {"http", "https"}:
                continue

            _logger.info("Saving URL %s", url)
            save_url = f"https://web.archive.org/save/{url}"
            async with lock:
                if i > 0:
                    await asyncio.sleep(sleep.total_seconds())

                async with session.get(save_url) as response:
                    for rel, params in response.links.items():
                        if rel == "memento":
                            saved_urls[url] = params["url"]

    return saved_urls



"""
Tests for ``nefelibata.assistants.mirror_images``.
"""
# pylint: disable=invalid-name, redefined-outer-name

from pathlib import Path
from typing import Dict

import pytest
from aiohttp import ClientSession
from pytest_mock import MockerFixture

from nefelibata.assistants.mirror_images import (
    MirrorImagesAssistant,
    add_exif,
    download_image,
    extract_images,
    get_filename,
    get_resource_extension,
    is_local,
)
from nefelibata.config import Config
from nefelibata.post import Post


@pytest.mark.asyncio
async def test_assistant(
    mocker: MockerFixture,
    root: Path,
    config: Config,
    post: Post,
) -> None:
    """
    Test the assistant.
    """
    mocker.patch(
        "nefelibata.assistants.mirror_images.download_image",
        return_value=mocker.AsyncMock(),
    )

    assistant = MirrorImagesAssistant(root, config)

    post.content = """
I have 2 images:

- ![First image](https://example.com/photo.jpg)
- ![Second image](img/logo.png)
    """

    await assistant.process_post(post)
    await assistant.process_post(post)

    async def modify_replacements(  # pylint: disable=too-many-arguments, unused-argument
        session: ClientSession,
        url: str,
        title: str,
        post: Post,
        mirror: Path,
        replacements: Dict[str, str],
    ) -> None:
        """
        A dummy function to modify ``replacements``.
        """
        replacements["hello"] = "world"

    mocker.patch(
        "nefelibata.assistants.mirror_images.download_image",
        modify_replacements,
    )

    await assistant.process_post(post)


def test_extract_images() -> None:
    """
    Test ``extract_images``.
    """

    assert list(extract_images("No images here, move along")) == []

    content = """
This is a long document with ![A title](https://example.com/photo.jpg).
    """
    assert list(extract_images(content)) == [
        ("https://example.com/photo.jpg", "A title"),
    ]


def test_is_local() -> None:
    """
    Test ``is_local``.
    """
    assert not is_local("https://example.com/photo.jpg")
    assert is_local("img/logo.png")


@pytest.mark.asyncio
async def test_get_resource_extension(mocker: MockerFixture) -> None:
    """
    Test ``get_resource_extension``.
    """
    session = mocker.MagicMock()
    session.head.return_value.__aenter__.return_value.headers = {
        "content-type": "image/jpeg",
    }

    extension = await get_resource_extension(session, "https://example.com/photo.jpg")
    assert extension == ".jpg"


def test_get_filename() -> None:
    """
    Test ``get_filename``.
    """
    assert (
        get_filename("https://example.com/photo.jpg", ".jpg")
        == "39c18449612764d8d87663444426c037.jpg"
    )


def test_add_exif(mocker: MockerFixture) -> None:
    """
    Test ``add_exif``.
    """
    BytesIO = mocker.patch("nefelibata.assistants.mirror_images.BytesIO")
    Image = mocker.patch("nefelibata.assistants.mirror_images.Image")
    piexif = mocker.patch("nefelibata.assistants.mirror_images.piexif")

    buf = BytesIO()
    add_exif(buf, "https://example.com/photo.jpg", "A title")

    piexif.dump.assert_called_with(
        {
            "0th": {
                piexif.ImageIFD.Copyright: "https://example.com/photo.jpg",
                piexif.ImageIFD.ImageDescription: "A title",
            },
        },
    )
    Image.open.return_value.save.assert_called_with(
        BytesIO.return_value,
        "jpeg",
        exif=piexif.dump.return_value,
    )


@pytest.mark.asyncio
async def test_download_image(mocker: MockerFixture, post: Post) -> None:
    """
    Test ``download_image``.
    """
    # mock network calls
    session = mocker.MagicMock()
    session.head.return_value.__aenter__.return_value.headers = {
        "content-type": "image/jpeg",
    }
    response = session.get.return_value.__aenter__.return_value = mocker.MagicMock()
    response.content.iter_chunked.return_value.__aiter__.return_value = [
        b"hello",
        b" ",
        b"world",
    ]

    # mock image manipulation
    add_exif = mocker.patch("nefelibata.assistants.mirror_images.add_exif")
    add_exif.return_value.getvalue.return_value = b"Hello"
    BytesIO = mocker.patch("nefelibata.assistants.mirror_images.BytesIO")

    _logger = mocker.patch("nefelibata.assistants.mirror_images._logger")

    mirror = post.path.parent / "img"
    mirror.mkdir()
    replacements: Dict[str, str] = {}

    # simple call
    await download_image(
        session,
        "https://example.com/photo.jpg",
        "A title",
        post,
        mirror,
        replacements,
    )
    assert (mirror / "39c18449612764d8d87663444426c037.jpg").exists()
    add_exif.assert_called_with(
        BytesIO.return_value,
        "https://example.com/photo.jpg",
        "A title",
    )
    assert replacements == {
        "https://example.com/photo.jpg": "img/39c18449612764d8d87663444426c037.jpg",
    }
    _logger.info.assert_called_with(
        "Downloading image from %s",
        "https://example.com/photo.jpg",
    )

    # call again
    await download_image(
        session,
        "https://example.com/photo.jpg",
        "A title",
        post,
        mirror,
        replacements,
    )
    _logger.debug.assert_called_with("Image already mirrored")

    # no EXIF
    session.head.return_value.__aenter__.return_value.headers = {
        "content-type": "image/png",
    }
    BytesIO.return_value.getvalue.return_value = b"Hello"
    add_exif.reset_mock()
    await download_image(
        session,
        "https://example.com/logo.png",
        "A title",
        post,
        mirror,
        replacements,
    )
    add_exif.assert_not_called()



"""
Assistant for saving external links in https://archive.org/.
"""

import logging
from typing import Dict

from nefelibata.announcers.base import Scope
from nefelibata.assistants.base import Assistant
from nefelibata.post import Post, extract_links
from nefelibata.utils import archive_urls

_logger = logging.getLogger(__name__)


class ArchiveLinksAssistant(Assistant):

    """
    Assistant for saving external links in https://archive.org/.
    """

    name = "saved_links"
    scopes = [Scope.POST]

    async def get_post_metadata(self, post: Post) -> Dict[str, str]:
        saved_urls = await archive_urls(extract_links(post))
        return {str(k): str(v) for k, v in saved_urls.items()}



"""
Test ``nefelibata.cli.new``.
"""
from pathlib import Path

import pytest
from pytest_mock import MockerFixture

from nefelibata.cli import new
from nefelibata.config import Config


@pytest.mark.asyncio
async def test_run(
    monkeypatch: pytest.MonkeyPatch,
    root: Path,
) -> None:
    """
    Test ``new``.
    """
    monkeypatch.setenv("EDITOR", "")

    await new.run(root, "A new post")

    filepath = root / "posts/a_new_post/index.mkd"
    with open(filepath, encoding="utf-8") as input_:
        content = input_.read()
    assert content == "subject: A new post\nsummary: \nkeywords: \n\n\n"

    with pytest.raises(IOError) as excinfo:
        await new.run(root, "A new post")
    assert str(excinfo.value) == "Directory already exists!"


@pytest.mark.asyncio
async def test_run_with_editor(
    monkeypatch: pytest.MonkeyPatch,
    mocker: MockerFixture,
    root: Path,
) -> None:
    """
    Test that the editor will be called if set.
    """
    monkeypatch.setenv("EDITOR", "vim")
    call = mocker.patch("nefelibata.cli.new.call")

    await new.run(root, "A new post")

    call.assert_called_with(["vim", root / "posts/a_new_post/index.mkd"])


@pytest.mark.asyncio
async def test_run_with_type(
    monkeypatch: pytest.MonkeyPatch,
    mocker: MockerFixture,
    root: Path,
    config: Config,
) -> None:
    """
    Test creating a new post with a valid type.
    """
    monkeypatch.setenv("EDITOR", "")
    config.templates = {
        "book": ["title", "author"],
    }
    mocker.patch("nefelibata.cli.new.get_config", return_value=config)

    await new.run(root, "A book I liked", "book")

    filepath = root / "posts/a_book_i_liked/index.mkd"
    with open(filepath, encoding="utf-8") as input_:
        content = input_.read()
    assert content == (
        "subject: A book I liked\n"
        "summary: \n"
        "keywords: \n"
        "type: book\n"
        "book-title: \n"
        "book-author: \n"
        "\n\n"
    )


@pytest.mark.asyncio
async def test_run_with_invalid_type(
    monkeypatch: pytest.MonkeyPatch,
    root: Path,
    config: Config,  # pylint: disable=unused-argument
) -> None:
    """
    Test creating a new post with an invalid type.
    """
    monkeypatch.setenv("EDITOR", "")

    with pytest.raises(Exception) as excinfo:
        await new.run(root, "A book I liked", "book")
    assert str(excinfo.value) == "Invalid post type: book"



"""
Tests for ``nefelibata.assistants.base``.
"""

from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, List, Type

import pytest
from freezegun import freeze_time
from pytest_mock import MockerFixture

from nefelibata.announcers.base import Scope
from nefelibata.assistants.base import Assistant, get_assistants
from nefelibata.config import AssistantModel, Config
from nefelibata.post import Post

from ..conftest import MockEntryPoint


def make_dummy_assistant(class_name: str, scopes: List[Scope]) -> Type[Assistant]:
    """
    Make a dummy ``Assistant`` derived class.
    """
    return type(class_name, (Assistant,), {"scopes": scopes})


@pytest.mark.asyncio
async def test_assistant(root: Path, config: Config, post: Post) -> None:
    """
    Test the base assistant methods.
    """

    assistant = Assistant(root, config)

    await assistant.process_post(post)
    assert not (post.path.parent / ".yaml").exists()
    assert (await assistant.get_post_metadata(post)) == {}

    await assistant.process_site()
    assert not (root / ".yaml").exists()
    assert (await assistant.get_site_metadata()) == {}


@pytest.mark.asyncio
async def test_assistant_dummy(root: Path, config: Config, post: Post) -> None:
    """
    Test a simple assistant.
    """

    class DummyAssistant(Assistant):
        """
        A dummy assistant.
        """

        name = "dummy"

        async def get_post_metadata(self, post: Post) -> Dict[str, Any]:
            return {"hello": "world"}

        async def get_site_metadata(self) -> Dict[str, Any]:
            return {"hello": "world"}

    assistant = DummyAssistant(root, config)

    with freeze_time("2021-01-01T00:00:00Z"):
        await assistant.process_post(post)
        await assistant.process_site()
    assert (post.path.parent / "dummy.yaml").stat().st_mtime == datetime(
        2021,
        1,
        1,
        0,
        0,
        tzinfo=timezone.utc,
    ).timestamp()
    assert (root / "dummy.yaml").stat().st_mtime == datetime(
        2021,
        1,
        1,
        0,
        0,
        tzinfo=timezone.utc,
    ).timestamp()

    with freeze_time("2021-01-02T00:00:00Z"):
        await assistant.process_post(post)
        await assistant.process_site()
    assert (post.path.parent / "dummy.yaml").stat().st_mtime == datetime(
        2021,
        1,
        1,
        0,
        0,
        tzinfo=timezone.utc,
    ).timestamp()
    assert (root / "dummy.yaml").stat().st_mtime == datetime(
        2021,
        1,
        1,
        0,
        0,
        tzinfo=timezone.utc,
    ).timestamp()

    with freeze_time("2021-01-03T00:00:00Z"):
        await assistant.process_post(post, force=True)
        await assistant.process_site(force=True)
    assert (post.path.parent / "dummy.yaml").stat().st_mtime == datetime(
        2021,
        1,
        3,
        0,
        0,
        tzinfo=timezone.utc,
    ).timestamp()
    assert (root / "dummy.yaml").stat().st_mtime == datetime(
        2021,
        1,
        3,
        0,
        0,
        tzinfo=timezone.utc,
    ).timestamp()


def test_get_assistants(
    mocker: MockerFixture,
    make_entry_point: Type[MockEntryPoint],
    root: Path,
    config: Config,
) -> None:
    """
    Test ``get_assistants``.
    """
    DummyAssistant = make_dummy_assistant("DummyAssistant", [Scope.SITE])
    entry_points = [
        make_entry_point("dummy_assistant", DummyAssistant),
    ]
    mocker.patch(
        "nefelibata.assistants.base.iter_entry_points",
        return_value=entry_points,
    )

    config.assistants = {
        "dummy_assistant": AssistantModel(
            **{
                "plugin": "dummy_assistant",
            }
        ),
    }
    assistants = get_assistants(root, config, Scope.SITE)
    assert len(assistants) == 1

    assistants = get_assistants(root, config, Scope.POST)
    assert len(assistants) == 0



"""
Tests for ``nefelibata.assistants.playlist``.
"""

from datetime import datetime, timezone
from pathlib import Path

import pytest
from freezegun import freeze_time

from nefelibata.assistants.playlist import PlaylistAssistant
from nefelibata.config import Config
from nefelibata.enclosure import MP3Enclosure
from nefelibata.post import Post


@pytest.mark.asyncio
async def test_assistant(root: Path, config: Config, post: Post) -> None:
    """
    Test the assistant.
    """
    assistant = PlaylistAssistant(root, config, base_url="https://example.com/mp3s/")

    index_path = post.path.parent / "index.pls"

    await assistant.process_post(post)
    assert not index_path.exists()

    post.enclosures.append(
        MP3Enclosure(
            path=root / "posts/first/song.mp3",
            description='"A title" (2m3s) by An artist (An album, 2021)',
            type="audio/mpeg",
            length=666,
            href="first/song.mp3",
            title="A title",
            artist="An artist",
            album="An album",
            year=2021,
            duration=123,
            track=0,
        ),
    )

    with freeze_time("2021-01-01T00:00:00Z"):
        await assistant.process_post(post)
    assert (
        index_path.stat().st_mtime
        == datetime(2021, 1, 1, tzinfo=timezone.utc).timestamp()
    )

    with open(index_path, encoding="utf-8") as input_:
        content = input_.read()
    assert (
        content
        == """[playlist]

NumberOfEntries=1
Version=2

File1=https://example.com/mp3s/first/song.mp3
Title1="A title" (2m3s) by An artist (An album, 2021)
Length1=123.0

"""
    )

    with freeze_time("2021-01-02T00:00:00Z"):
        await assistant.process_post(post)
    assert (
        index_path.stat().st_mtime
        == datetime(2021, 1, 1, tzinfo=timezone.utc).timestamp()
    )

    with freeze_time("2021-01-03T00:00:00Z"):
        await assistant.process_post(post, force=True)
    assert (
        index_path.stat().st_mtime
        == datetime(2021, 1, 3, tzinfo=timezone.utc).timestamp()
    )



"""
Tests for ``nefelibata.announcers.archive_blog``.
"""

from datetime import datetime, timezone
from pathlib import Path

import pytest
from freezegun import freeze_time
from pytest_mock import MockerFixture
from yarl import URL

from nefelibata.announcers.archive_blog import ArchiveBlogAnnouncer
from nefelibata.announcers.base import Announcement
from nefelibata.builders.base import Builder
from nefelibata.config import Config
from nefelibata.post import Post


@pytest.mark.asyncio
async def test_announcer_announce_post(
    mocker: MockerFixture,
    root: Path,
    config: Config,
    post: Post,
) -> None:
    """
    Test the announcer saving a post.
    """
    gemini_builder = Builder(root, config, "gemini://example.com/", "gemini")
    gemini_builder.extension = ".gmi"
    html_builder = Builder(root, config, "https://example.com/", "www")
    html_builder.extension = ".html"

    mocker.patch(
        "nefelibata.announcers.archive_blog.archive_urls",
        return_value={},
    )

    announcer = ArchiveBlogAnnouncer(
        root,
        config,
        builders=[gemini_builder, html_builder],
    )

    announcement = await announcer.announce_post(post)
    assert announcement is None

    archive_urls = mocker.patch(
        "nefelibata.announcers.archive_blog.archive_urls",
        return_value={
            URL("https://nefelibata.readthedocs.io/"): URL(
                "https://web.archive.org/web/20211003154602/https://nefelibata.readthedocs.io/",
            ),
        },
    )

    with freeze_time("2021-01-01T00:00:00Z"):
        announcement = await announcer.announce_post(post)
    assert announcement == Announcement(
        url="https://web.archive.org/save/",
        timestamp=datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
        grace_seconds=0,
    )

    archive_urls.assert_called_with(
        [
            URL("gemini://example.com/first/index.gmi"),
            URL("https://example.com/first/index.html"),
        ],
    )


@pytest.mark.asyncio
async def test_announcer_announce_site(
    mocker: MockerFixture,
    root: Path,
    config: Config,
) -> None:
    """
    Test the announcer saving a blog.
    """
    gemini_builder = Builder(root, config, "gemini://example.com/", "gemini")
    gemini_builder.extension = ".gmi"
    html_builder = Builder(root, config, "https://example.com/", "www")
    html_builder.extension = ".html"

    mocker.patch(
        "nefelibata.announcers.archive_blog.archive_urls",
        return_value={},
    )

    announcer = ArchiveBlogAnnouncer(
        root,
        config,
        builders=[gemini_builder, html_builder],
    )

    announcement = await announcer.announce_site()
    assert announcement is None

    archive_urls = mocker.patch(
        "nefelibata.announcers.archive_blog.archive_urls",
        return_value={
            URL("https://example.com/"): URL(
                "https://web.archive.org/web/20211003154602/https://example.com/",
            ),
        },
    )

    with freeze_time("2021-01-01T00:00:00Z"):
        announcement = await announcer.announce_site()
    assert announcement == Announcement(
        url="https://web.archive.org/save/",
        timestamp=datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
        grace_seconds=0,
    )

    archive_urls.assert_called_with(
        [
            URL("gemini://example.com/"),
            URL("https://example.com/"),
        ],
    )



"""
Base class for announcers.
"""
# pylint: disable=too-few-public-methods, no-self-use

import logging
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional

from pkg_resources import iter_entry_points
from pydantic import BaseModel

from nefelibata.builders.base import Builder, get_builders
from nefelibata.config import Config
from nefelibata.post import Post

_logger = logging.getLogger(__name__)


class Scope(Enum):
    """
    The scope of a given announcer.

    Announcers can process a single post, the entire site, or both.
    """

    POST = "POST"
    SITE = "SITE"


class Announcement(BaseModel):
    """
    Model representing an announcement.
    """

    url: str
    timestamp: datetime
    grace_seconds: int = 0


class Author(BaseModel):
    """
    Model representing the author of an interaction.
    """

    name: str
    url: str
    avatar: Optional[str] = None
    note: Optional[str] = None


class Interaction(BaseModel):
    """
    Model representing an interaction (like, reply, etc.).
    """

    id: str

    name: str
    summary: Optional[str] = None
    content: Optional[str] = None

    published: Optional[datetime] = None
    updated: Optional[datetime] = None

    author: Optional[Author] = None
    url: str
    in_reply_to: Optional[str] = None

    type: Literal["reply", "mention", "backlink", "like"]


class Announcer:

    """
    A post announcer.

    Announcers run after and before builders. They run after builders in order
    to publish the post, and before the builders in order to collect replies and
    other interactions.
    """

    scopes: List[Scope] = []

    def __init__(
        self, root: Path, config: Config, builders: List[Builder], **kwargs: Any
    ):
        self.root = root
        self.config = config
        self.builders = builders
        self.kwargs = kwargs

    async def announce_post(self, post: Post) -> Optional[Announcement]:
        """
        Publish a post externally.
        """

    async def announce_site(self) -> Optional[Announcement]:
        """
        Publish a site externally.
        """

    # pylint: disable=unused-argument
    async def collect_post(self, post: Post) -> Dict[str, Interaction]:
        """
        Collect interactions on a post.
        """
        return {}

    async def collect_site(self) -> Dict[Path, Dict[str, Interaction]]:
        """
        Collect interactions on a site.
        """
        return {}


def get_announcers(
    root: Path,
    config: Config,
    scope: Optional[Scope] = None,
) -> Dict[str, Announcer]:
    """
    Return configured announcers.
    """
    builders = get_builders(root, config)

    classes = {
        announcer.name: announcer.load()
        for announcer in iter_entry_points("nefelibata.announcer")
    }

    announcers = {}
    for announcer_name, announcer_config in config.announcers.items():
        class_ = classes[announcer_config.plugin]

        # find all builders that the announcer should handle
        announcer_builders = [
            builders[builder_name]
            for builder_name, builder_config in config.builders.items()
            if announcer_name in builder_config.announce_on
        ]

        if scope is None or scope in class_.scopes:
            announcers[announcer_name] = class_(
                root, config, announcer_builders, **announcer_config.dict()
            )

    return announcers



"""
Base class for assistants.
"""
# pylint: disable=unused-argument, no-self-use

import logging
from pathlib import Path
from typing import Any, Dict, List, Optional

import yaml
from pkg_resources import iter_entry_points

from nefelibata.announcers.base import Scope
from nefelibata.config import Config
from nefelibata.post import Post

_logger = logging.getLogger(__name__)


class Assistant:

    """
    A post assistant.

    Assistants run before builders, getting extra metadata and storing it in YAML
    files alongside the post.
    """

    name = ""
    scopes: List[Scope] = []

    def __init__(self, root: Path, config: Config, **kwargs: Any):
        self.root = root
        self.config = config
        self.kwargs = kwargs

    async def process_post(self, post: Post, force: bool = False) -> None:
        """
        Pre-process a post before it's built.
        """
        path = post.path.parent / f"{self.name}.yaml"
        if path.exists() and not force:
            return

        metadata = await self.get_post_metadata(post)
        if not metadata:
            return

        with open(path, "w", encoding="utf-8") as output:
            output.write(yaml.dump(metadata))

    async def get_post_metadata(self, post: Post) -> Dict[str, Any]:
        """
        Compute the metadata for a given post.
        """
        return {}

    async def process_site(self, force: bool = False) -> None:
        """
        Pre-process a site before it's built.
        """
        path = self.root / f"{self.name}.yaml"
        if path.exists() and not force:
            return

        metadata = await self.get_site_metadata()
        if not metadata:
            return

        with open(path, "w", encoding="utf-8") as output:
            output.write(yaml.dump(metadata))

    async def get_site_metadata(self) -> Dict[str, Any]:
        """
        Compute the metadata for the site.
        """
        return {}


def get_assistants(
    root: Path,
    config: Config,
    scope: Optional[Scope] = None,
) -> Dict[str, Assistant]:
    """
    Return configured assistants.
    """
    classes = {
        assistant.name: assistant.load()
        for assistant in iter_entry_points("nefelibata.assistant")
    }

    assistants = {}
    for assistant_name, assistant_config in config.assistants.items():
        class_ = classes[assistant_config.plugin]

        if scope is None or scope in class_.scopes:
            assistants[assistant_name] = class_(root, config, **assistant_config.dict())

    return assistants



"""
A Nefelibata post (and associated functions).
"""

import asyncio
import operator
from datetime import datetime
from email.header import decode_header, make_header
from email.parser import Parser
from email.utils import formatdate, parsedate_to_datetime
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Set

import marko
from pydantic import BaseModel, PrivateAttr
from yarl import URL

from nefelibata.config import Config
from nefelibata.enclosure import Enclosure, get_enclosures
from nefelibata.utils import load_extra_metadata, split_header


class Post(BaseModel):  # pylint: disable=too-few-public-methods
    """
    Model representing a post.
    """

    # path to the post
    path: Path

    # the title of the post, stored in the "subject" header
    title: str

    # post creation timestamp, stored in the "date" header
    timestamp: datetime

    # all the headers from the post
    metadata: Dict[str, Any]

    # special metadata
    tags: Set[str]
    categories: Set[str]
    announcers: Set[str]

    enclosures: List[Enclosure]

    # a custom type will use a custom template
    type: str

    # relative URL to the post, without any extensions, eg ``first/index``
    url: str

    # the Markdown contents of the file
    content: str

    # add a lock to manipulate the post file
    _lock: asyncio.Lock = PrivateAttr(default_factory=asyncio.Lock)


def build_post(root: Path, config: Config, path: Path) -> Post:
    """
    Build a post from a file path.

    Each post is stored in a file as an email.
    """
    with open(path, encoding="utf-8") as input_:
        parsed = Parser().parse(input_)

    # set default for missing values
    required = {
        "subject": "No subject found",
        "date": formatdate(path.stat().st_mtime, localtime=True),
    }
    modified = False
    for header, default in required.items():
        if header not in parsed:
            parsed[header] = default
            modified = True

    metadata = {
        k: str(make_header(decode_header(v)))
        for k, v in parsed.items()
        if k not in required
    }
    type_ = metadata.get("type", "post")
    tags = split_header(metadata.get("keywords"))
    categories = {
        category_name
        for category_name, category_config in config.categories.items()
        if tags & set(category_config.tags)
    }

    if "announce-on" in metadata:
        announcers = split_header(metadata.get("announce-on"))
    else:
        announcers = set(config.announcers)
        parsed["announce-on"] = ", ".join(announcers)
        modified = True

    if modified:
        with open(path, "w", encoding="utf-8") as output:
            output.write(str(parsed))

    # load metadata from YAML files
    metadata.update(load_extra_metadata(path.parent))

    return Post(
        path=path,
        title=str(make_header(decode_header(parsed["subject"]))),
        timestamp=parsedate_to_datetime(parsed["date"]),
        metadata=metadata,
        tags=tags,
        categories=categories,
        announcers=announcers,
        enclosures=get_enclosures(root, path.parent),
        type=type_,
        url=str(path.relative_to(root / "posts").with_suffix("")),
        content=parsed.get_payload(decode=False),
    )


def get_posts(root: Path, config: Config, count: Optional[int] = None) -> List[Post]:
    """
    Return all the posts.

    Posts are sorted by timestamp in descending order.
    """
    paths = list((root / "posts").glob("**/*.mkd"))
    posts = sorted(
        (build_post(root, config, path) for path in paths),
        key=operator.attrgetter("timestamp"),
        reverse=True,
    )
    return posts[:count]  # a[:None] == a


def extract_links(post: Post) -> Iterator[URL]:
    """
    Extract all links from a post.

    This includes links in the content, as well as metadata.
    """
    for key, value in post.metadata.items():
        if key.endswith("-url"):
            yield URL(value)

    tree = marko.parse(post.content)
    queue = [tree]
    while queue:
        element = queue.pop(0)

        if isinstance(element, marko.inline.Link):
            yield URL(element.dest)
        elif hasattr(element, "children"):
            queue.extend(element.children)



"""
Tests for ``nefelibata.assistants.news``.
"""

import random
from pathlib import Path

import pytest
from freezegun import freeze_time
from pytest_mock import MockerFixture

from nefelibata.assistants.news import NewsAssistant
from nefelibata.config import Config
from nefelibata.post import Post


@pytest.mark.asyncio
async def test_assistant(
    mocker: MockerFixture,
    root: Path,
    config: Config,
    post: Post,
) -> None:
    """
    Test the assistant.
    """
    random.seed(42)

    get = mocker.patch("nefelibata.assistants.news.ClientSession.get")
    get.return_value.__aenter__.return_value.json.return_value = {
        "articles": [
            {"hello": "world"},
            {"goodbye": "world"},
        ],
    }

    assistant = NewsAssistant(root, config, "SECRET", "us")

    with freeze_time("2021-01-01T00:00:00Z"):
        response = await assistant.get_post_metadata(post)
    assert response == {"hello": "world"}

    with freeze_time("2021-01-03T00:00:00Z"):
        response = await assistant.get_post_metadata(post)
    assert response == {}



"""
Tests for ``nefelibata.assistants.rsvp_calendar``.
"""
# pylint: disable=redefined-outer-name

from datetime import datetime, timezone
from pathlib import Path

import pytest
from freezegun import freeze_time
from icalendar import Event
from pytest_mock import MockerFixture
from yarl import URL

from nefelibata.assistants.rsvp_calendar import (
    RSVPCalendarAssistant,
    fetch_event,
    update_calendar,
)
from nefelibata.config import Config
from nefelibata.post import Post


@pytest.mark.asyncio
async def test_fetch_event_no_events(mocker: MockerFixture) -> None:
    """
    Test ``fetch_event`` when no events are found.
    """
    _logger = mocker.patch("nefelibata.assistants.rsvp_calendar._logger")

    get = mocker.patch("nefelibata.assistants.rsvp_calendar.ClientSession.get")
    get.return_value.__aenter__.return_value.text.return_value = "<p>Hello!</p>"

    event = await fetch_event(URL("https://example.com/events"))
    assert event is None
    _logger.warning.assert_called_with(
        "No events found on %s",
        URL("https://example.com/events"),
    )


@pytest.mark.asyncio
async def test_fetch_event_multiple_events(mocker: MockerFixture) -> None:
    """
    Test ``fetch_event`` when multiple events are found.
    """
    _logger = mocker.patch("nefelibata.assistants.rsvp_calendar._logger")

    get = mocker.patch("nefelibata.assistants.rsvp_calendar.ClientSession.get")
    get.return_value.__aenter__.return_value.text.return_value = """
<div class="h-event">
  <h1 class="p-name">Microformats Meetup</h1>
  <p>From
    <time class="dt-start" datetime="2013-06-30 12:00">30<sup>th</sup> June 2013, 12:00</time>
    to <time class="dt-end" datetime="2013-06-30 18:00">18:00</time>
    at <span class="p-location">Some bar in SF</span></p>
  <p class="p-summary">Get together and discuss all things microformats-related.</p>
</div>

<div class="h-event">
  <h1 class="p-name">Microformats Meetup</h1>
  <p>From
    <time class="dt-start" datetime="2013-06-30 12:00">30<sup>th</sup> June 2013, 12:00</time>
    to <time class="dt-end" datetime="2013-06-30 18:00">18:00</time>
    at <span class="p-location">Some bar in SF</span></p>
  <p class="p-summary">Get together and discuss all things microformats-related.</p>
</div>
    """

    event = await fetch_event(URL("https://example.com/events"))
    assert event is None
    _logger.warning.assert_called_with(
        "Multiple events found on %s",
        URL("https://example.com/events"),
    )


@pytest.mark.asyncio
async def test_fetch_event_simple_event(mocker: MockerFixture) -> None:
    """
    Test ``fetch_event``.
    """
    get = mocker.patch("nefelibata.assistants.rsvp_calendar.ClientSession.get")
    get.return_value.__aenter__.return_value.text.return_value = """
<div class="h-event">
  <h1 class="p-name">Microformats Meetup</h1>
  <p>From
    <time class="dt-start" datetime="2013-06-30 12:00">30<sup>th</sup> June 2013, 12:00</time>
    to <time class="dt-end" datetime="2013-06-30 18:00">18:00</time>
    at <span class="p-location">Some bar in SF</span></p>
  <p class="p-summary">Get together and discuss all things microformats-related.</p>
</div>
    """

    with freeze_time("2021-01-01T00:00:00Z"):
        event = await fetch_event(URL("https://example.com/events"))
    assert (
        event.to_ical()
        == b"""BEGIN:VEVENT\r
SUMMARY:Microformats Meetup\r
DTSTART;VALUE=DATE-TIME:20130630T120000\r
DTEND;VALUE=DATE-TIME:20130630T180000\r
DTSTAMP;VALUE=DATE-TIME:20210101T000000Z\r
END:VEVENT\r
"""
    )


@pytest.mark.asyncio
async def test_fetch_event_full_event(mocker: MockerFixture) -> None:
    """
    Test ``fetch_event``.
    """
    get = mocker.patch("nefelibata.assistants.rsvp_calendar.ClientSession.get")
    get.return_value.__aenter__.return_value.text.return_value = """
<div class="h-event">
  <img src="logo.png" class="u-featured">
  <h1 class="p-name"><a href="https://example.com/events" class="u-url">Microformats Meetup</a></h1>
  <ul>
    <li><span class="p-category">indieweb</span></li>
    <li><span class="p-category">microformats</span></li>
  </ul>
  <p>From
    <time class="dt-start" datetime="2013-06-30 12:00">30<sup>th</sup> June 2013, 12:00</time>
    to <time class="dt-end" datetime="2013-06-30 18:00">18:00</time>
    at <span class="p-location">Some bar in SF</span></p>
  <p class="p-summary">Get together and discuss all things microformats-related.</p>
  <p class="e-content">This is the content.</p>
</div>
    """

    with freeze_time("2021-01-01T00:00:00Z"):
        event = await fetch_event(URL("https://example.com/events"))
    assert (
        event.to_ical()
        == b"""BEGIN:VEVENT\r
SUMMARY:Microformats Meetup\r
DTSTART;VALUE=DATE-TIME:20130630T120000\r
DTEND;VALUE=DATE-TIME:20130630T180000\r
DTSTAMP;VALUE=DATE-TIME:20210101T000000Z\r
ATTACH:https://example.com/logo.png\r
CATEGORIES:indieweb,microformats\r
DESCRIPTION:This is the content.\r
URL:https://example.com/events\r
END:VEVENT\r
"""
    )


def test_update_calendar(mocker: MockerFixture) -> None:
    """
    Test ``update_calendar``.
    """
    client = mocker.MagicMock()
    calendar = client.principal.return_value.calendar.return_value

    event = Event()
    event.add("summary", "A summary")
    event.add("dtstart", datetime(2021, 1, 1, 12, tzinfo=timezone.utc))
    event.add("dtend", datetime(2021, 1, 1, 14, tzinfo=timezone.utc))

    update_calendar(client, "Personal", event)

    calendar.save_event.assert_called_with(
        b"""BEGIN:VCALENDAR\r
VERSION:2.0\r
PRODID:-//Nefelibata Corp.//CalDAV Client//EN\r
X-WR-CALNAME:Personal\r
BEGIN:VEVENT\r
SUMMARY:A summary\r
DTSTART;TZID=UTC;VALUE=DATE-TIME:20210101T120000Z\r
DTEND;TZID=UTC;VALUE=DATE-TIME:20210101T140000Z\r
END:VEVENT\r
END:VCALENDAR\r
""",
    )


def test_update_calendar_create(mocker: MockerFixture) -> None:
    """
    Test ``update_calendar`` when the calendar doesn't exist.
    """
    client = mocker.MagicMock()
    client.principal.return_value.calendar.side_effect = Exception("Not found")
    mocker.patch(
        "nefelibata.assistants.rsvp_calendar.caldav.error.NotFoundError",
        Exception,
    )

    event = mocker.MagicMock()

    update_calendar(client, "Personal", event)

    client.principal.return_value.make_calendar.assert_called_with(name="Personal")


def test_update_calendar_update_event(mocker: MockerFixture) -> None:
    """
    Test ``update_calendar`` when the event already exists.
    """
    client = mocker.MagicMock()
    calendar = client.principal.return_value.calendar.return_value
    existing_event = mocker.MagicMock()
    calendar.date_search.return_value = [existing_event]

    event = Event()
    event.add("summary", "A summary")
    event.add("dtstart", datetime(2021, 1, 1, 12, tzinfo=timezone.utc))
    event.add("dtend", datetime(2021, 1, 1, 14, tzinfo=timezone.utc))
    event.add("url", "https://example.com/events")

    update_calendar(client, "Personal", event)
    existing_event.delete.assert_not_called()

    existing_event.vobject_instance.vevent.url.value = "https://example.com/events"
    update_calendar(client, "Personal", event)
    existing_event.delete.assert_called_with()


@pytest.mark.asyncio
async def test_assistant(
    mocker: MockerFixture,
    root: Path,
    config: Config,
    post: Post,
) -> None:
    """
    Test the assistant.
    """
    assistant = RSVPCalendarAssistant(
        root,
        config,
        "https://calendar.example.com/",
        "username",
        "password",
        "Personal",
    )

    # not an RSVP post
    metadata = await assistant.get_post_metadata(post)
    assert metadata == {}

    # RSVP post without h-event
    post.type = "rsvp"
    post.metadata["rsvp-url"] = "https://example.com/events"
    fetch_event = mocker.patch("nefelibata.assistants.rsvp_calendar.fetch_event")
    fetch_event.return_value = None
    update_calendar = mocker.patch(
        "nefelibata.assistants.rsvp_calendar.update_calendar",
    )

    metadata = await assistant.get_post_metadata(post)
    assert metadata == {"event": "https://example.com/events"}
    update_calendar.assert_not_called()

    # RSVP post with h-event
    fetch_event.return_value = True
    await assistant.get_post_metadata(post)
    update_calendar.assert_called_with(assistant.client, "Personal", True)



"""
A builder for Gemini (https://gemini.circumlunar.space/).
"""
# pylint: disable=unused-argument, no-self-use

import logging
from pathlib import Path
from typing import Any, Iterator

import marko
from marko.renderer import Renderer

from nefelibata.builders.base import Builder
from nefelibata.config import Config

_logger = logging.getLogger(__name__)


def extract_links(paragraph) -> Iterator[marko.inline.Link]:
    """
    Extract links from a paragraph.
    """
    queue = [paragraph]
    while queue:
        element = queue.pop(0)

        if isinstance(element, (marko.inline.Link, marko.inline.AutoLink)):
            yield element
        elif hasattr(element, "children"):
            queue.extend(element.children)


class GeminiRenderer(Renderer):

    """
    Render Markdown to Gemtext.
    """

    def render_paragraph(self, element) -> str:
        """
        Render a paragraph.

        Links are collected and displayed after the paragraph.
        """
        paragraph = self.render_children(element) + "\n"

        links = list(extract_links(element))
        if not links:
            return paragraph

        return (
            paragraph
            + "\n"
            + "\n".join(self.render_paragraph_link(link) for link in links)
            + "\n"
        )

    def render_paragraph_link(self, element) -> str:
        """
        Render links after a paragraph.
        """
        if element.title:
            return f"=> {element.dest} {element.title}"

        return f"=> {element.dest} {self.render_children(element)}"

    def render_link_ref_def(self, element) -> str:
        """
        Render a link reference definition.
        """
        return ""

    def render_list(self, element) -> str:
        """
        Render a list.
        """
        return self.render_children(element)

    def render_list_item(self, element) -> str:
        """
        Render a list item.
        """
        return "* " + self.render_children(element)

    def render_quote(self, element) -> str:
        """
        Render a blockquote.
        """
        return "\n".join(
            f"> {line}" if line else ""
            for line in self.render_children(element).split("\n")
        )

    def render_fenced_code(self, element) -> str:
        """
        Render code inside triple backticks.
        """
        return "```\n" + self.render_children(element) + "```\n"

    def render_heading(self, element) -> str:
        """
        Render heading.

        The maximum level is 3.
        """
        level = min(element.level, 3)
        return "#" * level + " " + self.render_children(element) + "\n"

    render_setext_heading = render_heading

    def render_image(self, element) -> str:
        """
        Render an image.
        """
        if element.title:
            return f"=> {element.dest} {element.title}"

        return f"=> {element.dest} {self.render_children(element)}"

    def render_blank_line(self, element) -> str:
        """
        Render a blank line.
        """
        return "\n"

    render_line_break = render_blank_line

    def render_raw_text(self, element) -> str:
        """
        Render raw text.
        """
        return element.children

    render_code_span = render_raw_text
    render_literal = render_raw_text
    render_inline_html = render_raw_text


class GeminiBuilder(Builder):

    """
    A builder for the Gemini protocol.
    """

    name = "gemini"
    label = "Gemini"
    extension = ".gmi"
    template_base = ""
    site_templates = ["index.gmi", "feed.gmi"]

    def __init__(  # pylint: disable=too-many-arguments
        self,
        root: Path,
        config: Config,
        home: str,
        path: str = "",
        links: str = "paragraph",
        **kwargs: Any,
    ):
        super().__init__(root, config, home, path, **kwargs)
        self.links = links

        self.setup()

    def render(self, content: str) -> str:
        gemini = marko.Markdown(renderer=GeminiRenderer)
        return gemini.convert(content).strip()



"""
Assistant for syncing events to a calendar.
"""

import logging
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import Any, Dict, Optional

import caldav
import dateutil.parser
import mf2py
from aiohttp import ClientSession
from caldav import DAVClient
from icalendar import Calendar, Event
from yarl import URL

from nefelibata.announcers.base import Scope
from nefelibata.assistants.base import Assistant
from nefelibata.config import Config
from nefelibata.post import Post

_logger = logging.getLogger(__name__)


async def fetch_event(url: URL) -> Optional[Event]:
    """
    Extract a event from a page.

    We assume the event is represented as an h-event in the URL.
    """
    async with ClientSession() as session:
        async with session.get(url) as response:
            content = await response.text()

        parser = mf2py.Parser(content)
        hevents = parser.to_dict(filter_by_type="h-event")

        if not hevents:
            _logger.warning("No events found on %s", url)
            return None
        if len(hevents) > 1:
            _logger.warning("Multiple events found on %s", url)
            return None
        hevent = hevents[0]

        event = Event()
        event.add("summary", hevent["properties"]["name"][0])
        event.add("dtstart", dateutil.parser.parse(hevent["properties"]["start"][0]))
        event.add("dtend", dateutil.parser.parse(hevent["properties"]["end"][0]))
        event.add("dtstamp", datetime.now(timezone.utc))

        if "url" in hevent["properties"]:
            event.add("url", hevent["properties"]["url"][0])

        if "content" in hevent["properties"]:
            event.add("description", hevent["properties"]["content"][0]["value"])

        if "category" in hevent["properties"]:
            event.add("categories", hevent["properties"]["category"])

        if "featured" in hevent["properties"]:
            attachment_url = url.join(URL(hevent["properties"]["featured"][0]))
            event.add("attach", attachment_url)

    return event


def update_calendar(client: DAVClient, calendar_name: str, event: Event) -> None:
    """
    Create or update an event in a calendar.
    """
    principal = client.principal()
    try:
        calendar = principal.calendar(name=calendar_name)
    except caldav.error.NotFoundError:
        calendar = principal.make_calendar(name=calendar_name)

    # delete existing event, identified by URL
    if "url" in event:
        # pad search, otherwise it will return nothing
        existing_events = calendar.date_search(
            start=event["dtstart"].dt.date(),
            end=event["dtend"].dt.date() + timedelta(days=1),
            expand=False,
        )
        for existing_event in existing_events:
            vevent = existing_event.vobject_instance.vevent
            if hasattr(vevent, "url") and vevent.url.value == event["url"]:
                _logger.info("Found existing event, deleting")
                existing_event.delete()

    # wrap in BEGIN:VCALENDAR
    component = Calendar()
    component.add("prodid", "-//Nefelibata Corp.//CalDAV Client//EN")
    component.add("version", "2.0")
    component.add("x-wr-calname", calendar_name)
    component.add_component(event)

    _logger.info("Creating event")
    calendar.save_event(component.to_ical())


class RSVPCalendarAssistant(Assistant):
    """
    An assistant that adds RSVP'd events to a Caldav calendar.
    """

    name = "rsvp_calendar"
    scopes = [Scope.POST]

    def __init__(  # pylint: disable=too-many-arguments
        self,
        root: Path,
        config: Config,
        url: str,
        username: str,
        password: str,
        calendar: str,
        **kwargs: Any
    ):
        super().__init__(root, config, **kwargs)

        self.client = DAVClient(url=url, username=username, password=password)
        self.calendar = calendar

    async def get_post_metadata(self, post: Post) -> Dict[str, Any]:
        if post.type != "rsvp":
            return {}

        url = post.metadata["rsvp-url"]
        event = await fetch_event(URL(url))
        if event:
            update_calendar(self.client, self.calendar, event)

        return {"event": url}



"""
Tests for ``nefelibata.assistants.reading_time``.
"""

from pathlib import Path

import pytest

from nefelibata.assistants.reading_time import ReadingTimeAssistant
from nefelibata.config import Config
from nefelibata.post import Post


@pytest.mark.asyncio
async def test_assistant(root: Path, config: Config, post: Post) -> None:
    """
    Test the assistant.
    """
    assistant = ReadingTimeAssistant(root, config)

    payload = await assistant.get_post_metadata(post)
    assert payload == {
        "images": 0,
        "total_minutes": 1,
        "total_seconds": 3.6226415094339623,
        "words": 16,
    }



"""
An HTML builder.
"""
import logging
from pathlib import Path
from typing import Any

import marko

from nefelibata.builders.base import Builder
from nefelibata.config import Config

_logger = logging.getLogger(__name__)


class HTMLBuilder(Builder):

    """
    A simple HTML builder.
    """

    name = "html"
    label = "HTML"
    extension = ".html"
    site_templates = ["index.html", "atom.xml"]

    def __init__(  # pylint: disable=too-many-arguments
        self,
        root: Path,
        config: Config,
        home: str,
        path: str = "",
        theme: str = "minimal",
        **kwargs: Any,
    ):
        super().__init__(root, config, home, path, **kwargs)
        self.theme = theme
        self.template_base = f"{theme}/src/"

    def setup(self) -> None:
        super().setup()

        # symlink CSS
        css_directory = self.root / "templates/builders/html" / self.theme / "src/css"
        target = self.root / "build" / self.path / "css"
        if css_directory.exists() and not target.exists():
            target.symlink_to(css_directory, target_is_directory=True)

    @staticmethod
    def render(content: str) -> str:
        markdown = marko.Markdown(extensions=["codehilite"])
        return markdown.convert(content)



"""
A Geminispace (gemini://geminispace.info) announcer.
"""
import logging
import re
import urllib.parse
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import Any, Dict, List, Optional

from aiogemini.client import Client
from aiogemini.security import TOFUContext
from yarl import URL

from nefelibata.announcers.base import Announcement, Announcer, Interaction, Scope
from nefelibata.builders.base import Builder
from nefelibata.config import Config
from nefelibata.post import Post

_logger = logging.getLogger(__name__)


class GeminispaceAnnouncer(Announcer):

    """
    A Geminispace (gemini://geminispace.info) announcer.
    """

    scopes = [Scope.POST, Scope.SITE]

    def __init__(
        self, root: Path, config: Config, builders: List[Builder], **kwargs: Any
    ):
        super().__init__(root, config, builders, **kwargs)

        self.client = Client(TOFUContext({}))

    async def announce_site(self) -> Optional[Announcement]:
        """
        Add capsule to Geminispace.
        """
        for builder in self.builders:
            if builder.home.scheme != "gemini":
                raise Exception(
                    "Geminispace announcer only works with `gemini://` builds",
                )

            capsule_url = urllib.parse.quote_plus(str(builder.home))
            url = f"gemini://geminispace.info/add-seed?{capsule_url}"

            _logger.info("Announcing capsule %s to Geminispace", capsule_url)
            await self.client.get(URL(url))

        return Announcement(
            url="gemini://geminispace.info/",
            timestamp=datetime.now(timezone.utc),
            grace_seconds=timedelta(days=365).total_seconds(),
        )

    async def collect_post(self, post: Post) -> Dict[str, Interaction]:
        """
        Collect backlinks for a given post.
        """
        interactions: Dict[str, Interaction] = {}

        for builder in self.builders:
            if builder.home.scheme != "gemini":
                raise Exception(
                    "Geminispace announcer only works with `gemini://` builds",
                )

            post_url = urllib.parse.quote_plus(str(builder.absolute_url(post)))
            url = f"gemini://geminispace.info/backlinks?{post_url}"

            response = await self.client.get(URL(url))
            payload = await response.read()
            content = payload.decode("utf-8")

            state = 0
            for line in content.split("\n"):
                if line.startswith("###"):
                    state = 1 if "cross-capsule backlinks" in line else 2
                if state == 1 and line.startswith("=>"):
                    _, url, name = re.split(r"\s+", line, 2)

                    id_ = f"backlink,{url}"
                    interactions[id_] = Interaction(
                        id=id_,
                        name=name,
                        url=url,
                        type="backlink",
                    )

        return interactions



"""
Build the blog.
"""
import asyncio
import logging
from collections import defaultdict
from pathlib import Path
from typing import Dict

import yaml

from nefelibata.announcers.base import Announcer, Interaction, Scope, get_announcers
from nefelibata.assistants.base import get_assistants
from nefelibata.builders.base import get_builders
from nefelibata.constants import INTERACTIONS_FILENAME
from nefelibata.post import Post, get_posts
from nefelibata.utils import dict_merge, get_config, load_yaml

_logger = logging.getLogger(__name__)


async def collect_post(
    post: Post,
    announcer: Announcer,
    post_interactions,
) -> None:
    """
    Collect site interactions using a given announcer.
    """
    interactions = await announcer.collect_post(post)
    post_interactions[post.path].update(interactions)


async def collect_site(
    announcer: Announcer,
    post_interactions,
) -> None:
    """
    Collect site interactions using a given announcer.
    """
    interactions = await announcer.collect_site()
    dict_merge(post_interactions, interactions)


async def save_interactions(
    post_directory: Path,
    interactions: Dict[str, Interaction],
) -> None:
    """
    Save new post interactions.
    """
    path = post_directory / INTERACTIONS_FILENAME
    current_interactions = load_yaml(path, Interaction)
    # recursive update
    dict_merge(current_interactions, interactions)
    with open(path, "w", encoding="utf-8") as output:
        return yaml.dump(
            {
                name: interaction.dict()
                for name, interaction in current_interactions.items()
            },
            output,
        )


async def run(  # pylint: disable=too-many-locals
    root: Path,
    force: bool = False,
) -> None:
    """
    Build blog from Markdown files and online interactions.
    """
    _logger.info("Building blog")

    config = get_config(root)
    _logger.debug(config)

    build = root / "build"
    if not build.exists():
        _logger.info("Creating `build/` directory")
        build.mkdir()

    posts = get_posts(root, config)

    # collect interactions from posts/site
    tasks = []
    post_interactions: Dict[Path, Dict[str, Interaction]] = defaultdict(dict)

    _logger.info("Collecting interactions from posts")
    announcers = get_announcers(root, config, Scope.POST)
    for post in posts:
        for name, announcer in announcers.items():
            if name in post.announcers:
                task = asyncio.create_task(
                    collect_post(post, announcer, post_interactions),
                )
                tasks.append(task)

    _logger.info("Collecting interactions from site")
    announcers = get_announcers(root, config, Scope.SITE)
    for announcer in announcers.values():
        task = asyncio.create_task(collect_site(announcer, post_interactions))
        tasks.append(task)

    await asyncio.gather(*tasks)

    # store new interactions
    tasks = []
    for path, interactions in post_interactions.items():
        task = asyncio.create_task(save_interactions(path.parent, interactions))
        tasks.append(task)

    await asyncio.gather(*tasks)

    # run assistants
    tasks = []

    _logger.info("Running post assistants")
    assistants = get_assistants(root, config, Scope.POST)
    for post in posts:
        for assistant in assistants.values():
            task = asyncio.create_task(assistant.process_post(post, force))
            tasks.append(task)

    _logger.info("Running site assistants")
    assistants = get_assistants(root, config, Scope.SITE)
    for assistant in assistants.values():
        task = asyncio.create_task(assistant.process_site(force))
        tasks.append(task)

    await asyncio.gather(*tasks)

    # build posts/site
    tasks = []
    builders = get_builders(root, config)

    _logger.info("Processing posts")
    for post in posts:
        for builder in builders.values():
            task = asyncio.create_task(builder.process_post(post, force))
            tasks.append(task)

    _logger.info("Processing site")
    for builder in builders.values():
        task = asyncio.create_task(builder.process_site(force))
        tasks.append(task)

    await asyncio.gather(*tasks)



"""
Fake objects to simplify testing.
"""

from pathlib import Path

POST_DATA = {
    "path": Path("/path/to/blog/posts/first/index.mkd"),
    "title": "This is your first post",
    "timestamp": "2021-01-01T00:00:00Z",
    "metadata": {
        "keywords": "welcome, blog",
        "summary": "Hello, world!",
    },
    "tags": {"welcome", "blog"},
    "categories": ["stem"],
    "announcers": [],
    "enclosures": [],
    "type": "post",
    "url": "first/index",
    "content": """# Welcome #

This is your first post. It should be written using Markdown.

Read more about [Nefelibata](https://nefelibata.readthedocs.io/).""",
}

POST_CONTENT = """subject: This is your first post
keywords: welcome, blog
summary: Hello, world!
# Welcome #

This is your first post. It should be written using Markdown.

Read more about [Nefelibata](https://nefelibata.readthedocs.io/)."""


CONFIG = {
    "title": "道&c.",
    "subtitle": "Musings about the path and other things",
    "author": {
        "name": "Beto Dealmeida",
        "url": "https://taoetc.org/",
        "email": "roberto@dealmeida.net",
        "note": "Este, sou eu",
    },
    "language": "en",
    "social": [
        {"title": "My page", "url": "https://example.com/user"},
    ],
    "categories": {
        "stem": {
            "label": "STEM",
            "description": "Science, technology, engineering, & math",
            "tags": ["blog", "programming"],
        },
    },
    "templates": {"short": []},
    "builders": {
        "builder": {
            "plugin": "builder",
            "announce-on": ["announcer"],
            "publish-to": ["publisher"],
            "home": "https://example.com/",
            "path": "generic",
        },
    },
    "assistants": {
        "assistant": {
            "plugin": "assistant",
        },
    },
    "announcers": {
        "announcer": {
            "plugin": "announcer",
        },
    },
    "publishers": {
        "publisher": {
            "plugin": "publisher",
        },
    },
}



"""
Base class for publishers.
"""

from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Iterator, Optional

from pkg_resources import iter_entry_points
from pydantic import BaseModel

from nefelibata.config import Config


class Publishing(BaseModel):  # pylint: disable=too-few-public-methods
    """
    Model for publishings.
    """

    timestamp: datetime


class Publisher:

    """
    A blog publisher.
    """

    name = ""

    def __init__(self, root: Path, config: Config, path: str, **kwargs: Any):
        self.root = root
        self.config = config
        self.path = path
        self.kwargs = kwargs

    async def publish(
        self,
        since: Optional[datetime] = None,
        force: bool = False,
    ) -> Optional[Publishing]:
        """
        Publish the site to a given location.
        """
        raise NotImplementedError("Subclasses MUST implement `publish`")

    def find_modified_files(
        self,
        force: bool,
        since: Optional[datetime],
    ) -> Iterator[Path]:
        """
        Return modified files since ``since``.
        """
        # convert to timestamp to compare with ``st_mtime``
        last_published = since.timestamp() if since else 0

        build = self.root / "build" / self.path
        queue = [build]
        while queue:
            current = queue.pop()
            for path in current.glob("*"):
                if path.is_dir():
                    queue.append(path)
                elif force or path.stat().st_mtime > last_published:
                    yield path


def get_publishers(root: Path, config: Config) -> Dict[str, Publisher]:
    """
    Return configured publishers.

    Publishers are defined for each build, eg:

        builders:
          ...
          gemini:
            publish-to:
              - vsftp

        publishers:
          vsftp:
            ...

    """
    classes = {
        entry_point.name: entry_point.load()
        for entry_point in iter_entry_points("nefelibata.publisher")
    }

    publishers = {}
    for builder_name, builder_config in config.builders.items():
        for publisher_name in builder_config.publish_to:
            publisher_config = config.publishers[publisher_name]
            class_ = classes[publisher_config.plugin]
            path = builder_config.path

            name = f"{builder_name} => {publisher_name}"
            publishers[name] = class_(root, config, path, **publisher_config.dict())

    return publishers



"""
Enclosures for media associated with posts.
"""

import mimetypes
import urllib.parse
from pathlib import Path
from typing import Dict, List, Type, Union

import piexif
from mutagen.mp3 import MP3
from pydantic import BaseModel


class Enclosure(BaseModel):
    """
    An enclosure.
    """

    path: Path
    description: str
    type: str
    length: int
    href: str


def get_pretty_duration(duration: float) -> str:
    """
    Return a nice duration from number of seconds.
    """
    units = ["h", "m", "s"]
    parts = []
    while units:
        value = int(duration) % 60 if units else duration
        parts.append((value, units.pop()))
        duration //= 60

    parts = parts[::-1]

    # remove trailing zeros
    while parts and parts[-1][0] == 0:
        parts.pop()

    # remove leading zeros
    while parts and parts[0][0] == 0:
        parts.pop(0)

    return "".join(f"{value}{unit}" for (value, unit) in parts) or "0s"


class MP3Enclosure(Enclosure):
    """
    An MP3 file.
    """

    type = "audio/mpeg"
    title: str
    artist: str
    album: str
    year: Union[int, str]
    duration: float
    track: int

    @classmethod
    def from_path(cls, root: Path, path: Path) -> "MP3Enclosure":
        """
        Build an enclosure from a path to an MP3.
        """
        length = path.stat().st_size
        href = urllib.parse.quote(str(path.relative_to(root / "posts")))

        mp3 = MP3(path)
        title = str(mp3.get("TIT2", path.stem))
        artist = str(mp3.get("TPE1", "Unknown artist"))
        album = str(mp3.get("TALB", "Unknown album"))
        year = int(str(mp3.get("TDRC", 0))) or "Unknown year"
        duration = mp3.info.length
        track = int(str(mp3.get("TRCK"))) if "TRCK" in mp3 else 0

        pretty_duration = get_pretty_duration(duration)
        description = f'"{title}" ({pretty_duration}) by {artist} ({album}, {year})'

        return cls(
            path=path,
            description=description,
            length=length,
            href=href,
            title=title,
            artist=artist,
            album=album,
            year=year,
            duration=duration,
            track=track,
        )


class ImageEnclosure(Enclosure):
    """
    An image.
    """

    path: Path
    description: str
    type: str
    length: int
    href: str

    @classmethod
    def from_path(cls, root: Path, path: Path) -> "ImageEnclosure":
        """
        Build an enclosure from a path to an image.
        """
        mimetype, _ = mimetypes.guess_type(path)
        length = path.stat().st_size
        href = urllib.parse.quote(str(path.relative_to(root / "posts")))

        if mimetype == "image/jpeg":
            exif = piexif.load(str(path))
            description = (
                exif["0th"][piexif.ImageIFD.ImageDescription] or f"Image {path.name}"
            )
        else:
            description = f"Image {path.name}"

        return cls(
            path=path,
            description=description,
            type=mimetype,
            length=length,
            href=href,
        )


mimetype_map: Dict[str, Type[Enclosure]] = {
    "audio/mpeg": MP3Enclosure,
    "image/jpeg": ImageEnclosure,
    "image/png": ImageEnclosure,
}


def get_enclosures(root: Path, post_directory: Path) -> List[Enclosure]:
    """
    Find all enclosures in a given post.
    """
    enclosures = []
    for file_path in post_directory.glob("**/*"):
        mimetype, _ = mimetypes.guess_type(file_path)
        class_ = mimetype_map.get(mimetype) if mimetype else None
        if class_:
            enclosures.append(class_.from_path(root, file_path))

    return enclosures



"""
Test ``nefelibata.cli.build``.
"""
# pylint: disable=invalid-name

from pathlib import Path

import pytest
from pytest_mock import MockerFixture

from nefelibata.cli import build
from nefelibata.post import Post


@pytest.mark.asyncio
async def test_run(
    mocker: MockerFixture,
    root: Path,
    post: Post,
) -> None:
    """
    Test ``run``.
    """
    assistant = mocker.MagicMock()
    assistant.process_post = mocker.AsyncMock()
    assistant.process_site = mocker.AsyncMock()

    builder = mocker.MagicMock()
    builder.process_post = mocker.AsyncMock()
    builder.process_site = mocker.AsyncMock()

    announcer1 = mocker.MagicMock()
    announcer1.collect_post = mocker.AsyncMock(return_value={})
    announcer1.collect_site = mocker.AsyncMock(return_value={})

    announcer2 = mocker.MagicMock()
    announcer2.collect_site = mocker.AsyncMock(return_value={})
    announcer2.collect_site.return_value = {}

    mocker.patch(
        "nefelibata.cli.build.get_announcers",
        return_value={"announcer1": announcer1, "announcer2": announcer2},
    )
    mocker.patch(
        "nefelibata.cli.build.get_assistants",
        return_value={"assistant": assistant},
    )
    mocker.patch("nefelibata.cli.build.get_builders", return_value={"builder": builder})
    mocker.patch("nefelibata.cli.build.get_config")
    mocker.patch("nefelibata.cli.build.get_posts", return_value=[post])

    _logger = mocker.patch("nefelibata.cli.build._logger")

    post.announcers = {"announcer1"}

    await build.run(root)

    assistant.process_post.assert_called_with(post, False)
    assistant.process_site.assert_called_with(False)
    builder.process_post.assert_called_with(post, False)
    builder.process_site.assert_called_with(False)
    announcer1.collect_post.assert_called_with(post)
    announcer1.collect_site.assert_called_with()
    announcer2.collect_post.assert_not_called()
    announcer2.collect_site.assert_called_with()
    _logger.info.assert_has_calls(
        [
            mocker.call("Building blog"),
            mocker.call("Creating `build/` directory"),
            mocker.call("Collecting interactions from posts"),
            mocker.call("Collecting interactions from site"),
            mocker.call("Running post assistants"),
            mocker.call("Running site assistants"),
            mocker.call("Processing posts"),
            mocker.call("Processing site"),
        ],
    )

    _logger.reset_mock()

    await build.run(root)
    _logger.info.assert_has_calls(
        [
            mocker.call("Processing posts"),
            mocker.call("Processing site"),
        ],
    )



"""
Tests for ``nefelibata.publishers.command``.
"""
# pylint: disable=invalid-name
from datetime import datetime, timezone
from pathlib import Path

import pytest
from freezegun import freeze_time
from pyfakefs.fake_filesystem import FakeFilesystem
from pytest_mock import MockerFixture

from nefelibata.config import Config
from nefelibata.publishers.command import CommandPublisher


@pytest.mark.asyncio
async def test_publish(
    mocker: MockerFixture,
    fs: FakeFilesystem,
    root: Path,
    config: Config,
) -> None:
    """
    Test ``publish``.
    """
    mocker.patch.object(
        CommandPublisher,
        "find_modified_files",
        return_value=iter(
            [
                root / "build/generic/one/index.html",
                root / "build/generic/one/photo.jpg",
                root / "build/generic/one/blob",
            ],
        ),
    )
    subprocess = mocker.patch("nefelibata.publishers.command.subprocess")

    fs.create_file(root / "build/generic/one", contents="Hello, world!")

    publisher = CommandPublisher(
        root,
        config,
        "generic",
        ["touch $path", "cp $path dir/"],
        ["notify.py"],
    )

    with freeze_time("2021-01-01T00:00:00Z"):
        publishing = await publisher.publish()
    assert publishing.dict() == {
        "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
    }
    print(subprocess.run.mock_calls)
    subprocess.run.assert_has_calls(
        [
            mocker.call(
                "touch $path",
                shell=True,
                check=True,
                env={
                    "path": Path("one/index.html"),
                },
            ),
            mocker.call(
                "cp $path dir/",
                shell=True,
                check=True,
                env={
                    "path": Path("one/index.html"),
                },
            ),
            mocker.call(
                "touch $path",
                shell=True,
                check=True,
                env={
                    "path": Path("one/photo.jpg"),
                },
            ),
            mocker.call(
                "cp $path dir/",
                shell=True,
                check=True,
                env={
                    "path": Path("one/photo.jpg"),
                },
            ),
            mocker.call(
                "touch $path",
                shell=True,
                check=True,
                env={
                    "path": Path("one/blob"),
                },
            ),
            mocker.call(
                "cp $path dir/",
                shell=True,
                check=True,
                env={
                    "path": Path("one/blob"),
                },
            ),
            mocker.call(
                "notify.py",
                shell=True,
                check=True,
            ),
        ],
    )


@pytest.mark.asyncio
async def test_publish_no_modified_files(
    mocker: MockerFixture,
    root: Path,
    config: Config,
) -> None:
    """
    Test ``publish`` when no files have been modified.
    """
    mocker.patch.object(
        CommandPublisher,
        "find_modified_files",
        return_value=iter([]),
    )
    mocker.patch("nefelibata.publishers.command.subprocess")

    publisher = CommandPublisher(
        root,
        config,
        "generic",
        ["touch $path", "cp $path dir/"],
        ["notify.py"],
    )

    with freeze_time("2021-01-01T00:00:00Z"):
        publishing = await publisher.publish()
    assert publishing is None



"""
Tests for ``nefelibata.assistants.current_weather``.
"""

from pathlib import Path

import pytest
from freezegun import freeze_time
from pytest_mock import MockerFixture

from nefelibata.assistants.current_weather import CurrentWeatherAssistant
from nefelibata.config import Config
from nefelibata.post import Post


@pytest.mark.asyncio
async def test_assistant(
    mocker: MockerFixture,
    root: Path,
    config: Config,
    post: Post,
) -> None:
    """
    Test the assistant.
    """
    get = mocker.patch("nefelibata.assistants.current_weather.ClientSession.get")
    get.return_value.__aenter__.return_value.json.return_value = {"hello": "world"}

    assistant = CurrentWeatherAssistant(root, config)

    with freeze_time("2021-01-01T00:00:00Z"):
        response = await assistant.get_post_metadata(post)
    assert response == {"hello": "world"}

    with freeze_time("2021-01-03T00:00:00Z"):
        response = await assistant.get_post_metadata(post)
    assert response == {}



"""
Assistant for fetching a random news headline.
"""

import logging
import random
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import Any, Dict

from aiohttp import ClientSession

from nefelibata.announcers.base import Scope
from nefelibata.assistants.base import Assistant
from nefelibata.config import Config
from nefelibata.post import Post

_logger = logging.getLogger(__name__)


DEFAULT_MAX_AGE = timedelta(days=1)


class NewsAssistant(Assistant):
    """
    Assistant for fetching a random news headline.
    """

    name = "news"
    scopes = [Scope.POST]

    def __init__(  # pylint: disable=too-many-arguments
        self,
        root: Path,
        config: Config,
        api_key: str,
        country: str = "us",
        max_age: timedelta = DEFAULT_MAX_AGE,
        **kwargs: Any
    ):
        super().__init__(root, config, **kwargs)

        self.api_key = api_key
        self.country = country
        self.max_age = max_age

    async def get_post_metadata(self, post: Post) -> Dict[str, Any]:
        if datetime.now(tz=timezone.utc) - post.timestamp > self.max_age:
            return {}

        params = {
            "country": self.country,
            "apiKey": self.api_key,
        }

        async with ClientSession() as session:
            _logger.info("Fetching a random news headline")
            async with session.get(
                "https://newsapi.org/v2/top-headlines",
                params=params,
            ) as response:
                payload = await response.json()
                articles = payload["articles"]
                return random.choice(articles)



"""
Tests for ``nefelibata.announcers.campcom``.
"""
# pylint: disable=invalid-name

from datetime import datetime, timezone
from pathlib import Path

import pytest
from freezegun import freeze_time
from pytest_mock import MockerFixture
from yarl import URL

from nefelibata.announcers.base import Interaction
from nefelibata.announcers.geminispace import GeminispaceAnnouncer
from nefelibata.builders.base import Builder
from nefelibata.config import Config
from nefelibata.post import Post


@pytest.mark.asyncio
async def test_announcer_announce(
    mocker: MockerFixture,
    root: Path,
    config: Config,
    post: Post,
) -> None:
    """
    Test announcing sites and posts.
    """
    gemini_builder = Builder(root, config, "gemini://example.com/", "gemini")
    html_builder = Builder(root, config, "https://example.com/", "www")

    Client = mocker.patch("nefelibata.announcers.geminispace.Client")
    Client.return_value.get = mocker.AsyncMock()

    announcer = GeminispaceAnnouncer(root, config, builders=[gemini_builder])

    # post announcement is no-op
    announcement = await announcer.announce_post(post)
    assert announcement is None

    with freeze_time("2021-01-01T00:00:00Z"):
        announcement = await announcer.announce_site()
    assert announcement.dict() == {
        "url": "gemini://geminispace.info/",
        "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
        "grace_seconds": 31536000,
    }
    assert Client.return_value.get.called_with(
        URL("gemini://geminipace.info/add-seed?gemini://example.com/feed"),
    )

    announcer = GeminispaceAnnouncer(root, config, builders=[html_builder])
    with pytest.raises(Exception) as excinfo:
        await announcer.announce_site()
    assert (
        str(excinfo.value) == "Geminispace announcer only works with `gemini://` builds"
    )


@pytest.mark.asyncio
async def test_announcer_collect(
    mocker: MockerFixture,
    root: Path,
    config: Config,
    post: Post,
) -> None:
    """
    Test collecting interactions on posts and sites.
    """
    gemini_builder = Builder(root, config, "gemini://example.com/", "gemini")
    html_builder = Builder(root, config, "https://example.com/", "www")

    Client = mocker.patch("nefelibata.announcers.geminispace.Client")
    Client.return_value.get = mocker.AsyncMock()
    Client.return_value.get.return_value.read.return_value = b"""
### 3 cross-capsule backlinks

=> gemini://example.com/reply.gmi Re: This is your first post
=> gemini://example.com/another-reply.gmi Re: This is your first post
    """

    announcer = GeminispaceAnnouncer(root, config, builders=[gemini_builder])

    # site collect is no-op
    interactions = await announcer.collect_site()
    assert interactions == {}

    interactions = await announcer.collect_post(post)
    assert interactions == {
        "backlink,gemini://example.com/reply.gmi": Interaction(
            id="backlink,gemini://example.com/reply.gmi",
            name="Re: This is your first post",
            url="gemini://example.com/reply.gmi",
            type="backlink",
            timestamp=None,
        ),
        "backlink,gemini://example.com/another-reply.gmi": Interaction(
            id="backlink,gemini://example.com/another-reply.gmi",
            name="Re: This is your first post",
            url="gemini://example.com/another-reply.gmi",
            type="backlink",
            timestamp=None,
        ),
    }

    announcer = GeminispaceAnnouncer(root, config, builders=[html_builder])
    with pytest.raises(Exception) as excinfo:
        await announcer.collect_post(post)
    assert (
        str(excinfo.value) == "Geminispace announcer only works with `gemini://` builds"
    )



"""
Assistant for fetching the current weather.
"""

import logging
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import Any, Dict

from aiohttp import ClientSession

from nefelibata.announcers.base import Scope
from nefelibata.assistants.base import Assistant
from nefelibata.config import Config
from nefelibata.post import Post

_logger = logging.getLogger(__name__)


DEFAULT_MAX_AGE = timedelta(days=1)


class CurrentWeatherAssistant(Assistant):
    """
    An assistant for fetching the current weather.
    """

    name = "current_weather"
    scopes = [Scope.POST]

    def __init__(
        self,
        root: Path,
        config: Config,
        max_age: timedelta = DEFAULT_MAX_AGE,
        **kwargs: Any
    ):
        super().__init__(root, config, **kwargs)

        self.max_age = max_age

    async def get_post_metadata(self, post: Post) -> Dict[str, Any]:
        if datetime.now(tz=timezone.utc) - post.timestamp > self.max_age:
            return {}

        async with ClientSession() as session:
            _logger.info("Fetching current weather information")
            async with session.get("https://wttr.in/?format=j1&m") as response:
                return await response.json()



"""
Tests for ``nefelibata.announcers.antenna``.
"""
# pylint: disable=invalid-name

import ssl
from datetime import datetime, timezone
from pathlib import Path

import pytest
from freezegun import freeze_time
from pytest_mock import MockerFixture
from yarl import URL

from nefelibata.announcers.antenna import AntennaAnnouncer
from nefelibata.announcers.base import Interaction
from nefelibata.builders.base import Builder
from nefelibata.config import Config
from nefelibata.post import Post


@pytest.mark.asyncio
async def test_announcer_announce(
    mocker: MockerFixture,
    root: Path,
    config: Config,
    post: Post,
) -> None:
    """
    Test announcing sites and posts.
    """
    gemini_builder = Builder(root, config, "gemini://example.com/", "gemini")
    html_builder = Builder(root, config, "https://example.com/", "www")

    Client = mocker.patch("nefelibata.announcers.gemlog.Client")
    Client.return_value.get = mocker.AsyncMock()

    announcer = AntennaAnnouncer(root, config, builders=[gemini_builder])

    # post announcement is no-op
    announcement = await announcer.announce_post(post)
    assert announcement is None

    with freeze_time("2021-01-01T00:00:00Z"):
        announcement = await announcer.announce_site()
    assert announcement.dict() == {
        "url": "gemini://warmedal.se/~antenna/",
        "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
        "grace_seconds": 0,
    }
    assert Client.return_value.get.called_with(
        URL("gemini://warmedal.se/~antenna/submit?gemini://example.com/feed"),
    )

    announcer = AntennaAnnouncer(root, config, builders=[html_builder])
    with pytest.raises(Exception) as excinfo:
        await announcer.announce_site()
    assert str(excinfo.value) == "Antenna announcer only works with `gemini://` builds"


@pytest.mark.asyncio
async def test_announcer_collect(
    mocker: MockerFixture,
    root: Path,
    config: Config,
    post: Post,
) -> None:
    """
    Test collecting interactions on posts and sites.
    """
    builder = Builder(root, config, "gemini://example.com/", "gemini")

    Client = mocker.patch("nefelibata.announcers.gemlog.Client")
    Client.return_value.get = mocker.AsyncMock()
    Client.return_value.get.return_value.read.side_effect = [
        b"""
=> gemini://example.com/reply.gmi Re: This is your first post
    """,
        b"""
=> gemini://example.com/first/index.gmi
    """,
    ]

    announcer = AntennaAnnouncer(root, config, builders=[builder])

    # post collect is no-op
    interactions = await announcer.collect_post(post)
    assert interactions == {}

    interactions = await announcer.collect_site()
    assert interactions == {
        Path("/path/to/blog/posts/first/index.mkd"): {
            "reply,gemini://example.com/reply.gmi": Interaction(
                id="reply,gemini://example.com/reply.gmi",
                name="Re: This is your first post",
                url="gemini://example.com/reply.gmi",
                type="reply",
                timestamp=None,
            ),
        },
    }

    # test no backlink
    Client.return_value.get.return_value.read.side_effect = [
        b"""
=> gemini://example.com/reply.gmi Re: This is your first post
    """,
        b"",
    ]

    announcer = AntennaAnnouncer(root, config, builders=[builder])

    interactions = await announcer.collect_site()
    assert interactions == {}

    # test SSL error
    response = mocker.AsyncMock()
    response.read.return_value = b"""
=> gemini://example.com/reply.gmi Re: This is your first post
    """
    Client.return_value.get.side_effect = [
        response,
        ssl.SSLCertVerificationError("A wild error appears!"),
    ]

    announcer = AntennaAnnouncer(root, config, builders=[builder])

    interactions = await announcer.collect_site()
    assert interactions == {
        Path("/path/to/blog/posts/first/index.mkd"): {
            "reply,gemini://example.com/reply.gmi": Interaction(
                id="reply,gemini://example.com/reply.gmi",
                name="Re: This is your first post",
                url="gemini://example.com/reply.gmi",
                type="reply",
                timestamp=None,
            ),
        },
    }



"""
Tests for ``nefelibata.publishers.ftp``.
"""
# pylint: disable=invalid-name
import ftplib
from datetime import datetime, timezone
from pathlib import Path

import pytest
from freezegun import freeze_time
from pyfakefs.fake_filesystem import FakeFilesystem
from pytest_mock import MockerFixture

from nefelibata.config import Config
from nefelibata.publishers.ftp import FTPPublisher


@pytest.mark.asyncio
async def test_publish(
    mocker: MockerFixture,
    fs: FakeFilesystem,
    root: Path,
    config: Config,
) -> None:
    """
    Test ``publish``.
    """
    FTP = mocker.patch("nefelibata.publishers.ftp.FTP")
    ftp = FTP.return_value.__enter__.return_value
    mocker.patch.object(
        FTPPublisher,
        "find_modified_files",
        return_value=iter([root / "build/generic/one"]),
    )

    fs.create_file(root / "build/generic/one", contents="Hello, world!")

    publisher = FTPPublisher(
        root,
        config,
        "generic",
        "ftp.example.com",
        "username",
        "password",
    )

    await publisher.publish()

    FTP.assert_called_with("ftp.example.com", "username", "password")
    ftp.pwd.assert_called_with()
    ftp.storbinary.assert_called()  # second argument is an internal object


@pytest.mark.asyncio
async def test_publish_last_published(
    mocker: MockerFixture,
    fs: FakeFilesystem,
    root: Path,
    config: Config,
) -> None:
    """
    Test publishing when ``last_published`` is present.
    """
    FTP = mocker.patch("nefelibata.publishers.ftp.FTP")
    ftp = FTP.return_value.__enter__.return_value

    with freeze_time("2021-01-01T00:00:00Z"):
        fs.create_file(root / "build/generic/one", contents="Hello, world!")

    publisher = FTPPublisher(
        root,
        config,
        "generic",
        "ftp.example.com",
        "username",
        "password",
    )

    await publisher.publish(since=datetime(2021, 1, 2, tzinfo=timezone.utc))

    ftp.storbinary.assert_not_called()

    await publisher.publish(force=True)

    ftp.storbinary.assert_called()


@pytest.mark.asyncio
async def test_publish_tls(
    mocker: MockerFixture,
    fs: FakeFilesystem,
    root: Path,
    config: Config,
) -> None:
    """
    Test ``publish`` with TLS.
    """
    FTP_TLS = mocker.patch("nefelibata.publishers.ftp.FTP_TLS")
    ftp = FTP_TLS.return_value.__enter__.return_value
    mocker.patch.object(
        FTPPublisher,
        "find_modified_files",
        return_value=iter([root / "build/generic/one"]),
    )

    fs.create_file(root / "build/generic/one", contents="Hello, world!")

    publisher = FTPPublisher(
        root,
        config,
        "generic",
        "ftp.example.com",
        "username",
        "password",
        use_tls=True,
    )

    await publisher.publish()

    ftp.prot_p.assert_called()


@pytest.mark.asyncio
async def test_publish_with_basedir(
    mocker: MockerFixture,
    fs: FakeFilesystem,
    root: Path,
    config: Config,
) -> None:
    """
    Test ``publish`` when a base directory is specified.
    """
    FTP = mocker.patch("nefelibata.publishers.ftp.FTP")
    ftp = FTP.return_value.__enter__.return_value
    ftp.pwd.return_value = "/ftp/upload"
    mocker.patch.object(
        FTPPublisher,
        "find_modified_files",
        return_value=iter(
            [root / "build/generic/subdir1/one", root / "build/generic/subdir2/two"],
        ),
    )

    fs.create_file(root / "build/generic/subdir1/one", contents="Hello, world!")
    fs.create_file(root / "build/generic/subdir2/two", contents="Goodbye, world!")

    publisher = FTPPublisher(
        root,
        config,
        "generic",
        "ftp.example.com",
        "username",
        "password",
        "ftp/upload",
    )

    await publisher.publish()

    ftp.cwd.assert_has_calls(
        [
            mocker.call("ftp/upload"),
            mocker.call("subdir1"),
            mocker.call("/ftp/upload"),
            mocker.call("subdir2"),
        ],
    )


@pytest.mark.asyncio
async def test_publish_create_directory(
    mocker: MockerFixture,
    fs: FakeFilesystem,
    root: Path,
    config: Config,
) -> None:
    """
    Test ``publish`` when the directories need to be created.
    """
    FTP = mocker.patch("nefelibata.publishers.ftp.FTP")
    ftp = FTP.return_value.__enter__.return_value
    ftp.cwd.side_effect = [ftplib.error_perm, ""]
    mocker.patch.object(
        FTPPublisher,
        "find_modified_files",
        return_value=iter([root / "build/generic/subdir1/one"]),
    )

    fs.create_file(root / "build/generic/subdir1/one", contents="Hello, world!")

    publisher = FTPPublisher(
        root,
        config,
        "generic",
        "ftp.example.com",
        "username",
        "password",
    )

    await publisher.publish()

    ftp.mkd.assert_called_with("subdir1")
